"""Test legacy functions."""

from rimseval.compatibility import lion_eval


def test_extract_block():
    """Extract a block from some data."""
    top_char = "# start_at_top"
    bott_char = "# end_at_bottom"
    data = [
        "# something",
        top_char,
        "0",
        "1",
        "2",
        bott_char + " something else longer",
        "3",
        "4",
        "5",
    ]
    data_block_exp = ["0", "1", "2"]
    # bottom char set automatically
    assert lion_eval._extract_block(data, top_char=top_char) == data_block_exp



"""Utilities for tests."""

import numpy as np

from rimseval.data_io import CRDReader


def assert_crd_equal(crd1: CRDReader, crd2: CRDReader):
    """Test function to assert that two CRD files are equal.

    Test the content of the CRD file after reading to ensure that they are identical.

    :param crd1: First CRD file to compare.
    :param crd2: Second CRD file to compare.
    """
    np.testing.assert_equal(crd1.all_tofs, crd2.all_tofs)
    np.testing.assert_equal(crd1.ions_per_shot, crd2.ions_per_shot)
    np.testing.assert_equal(crd1.ions_to_tof_map, crd2.ions_to_tof_map)
    assert crd1.nof_ions == crd2.nof_ions
    assert crd1.nof_shots == crd2.nof_shots



"""Configuration file for the Sphinx documentation builder."""

import sys

sys.path.append("../")

# -- Project information -----------------------------------------------------

project = "RIMSEval"
author = "Reto Trappitsch"
copyright = f"2021, {author}"
version = "2.0.0"
release = version


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = ["sphinx.ext.autodoc", "sphinx_rtd_theme"]

# Add any paths that contain templates here, relative to this directory.
# templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The main toctree document.
master_doc = "index"

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "sphinx"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
html_theme = "sphinx_rtd_theme"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
# html_static_path = ["_static"]



"""Macro template."""

from rimseval.processor import CRDFileProcessor


def calc(crd: CRDFileProcessor) -> None:
    """Macro template.

    :param crd: CRD file processor that will be passed to the macro.
    """
    # your code goes here



"""Interactive integral and background selection using matplotlib's qtagg backend."""

import sys
from typing import List, Union

import matplotlib.colors as mcolors
import numpy as np
from PyQt6 import QtCore, QtWidgets


from rimseval.processor import CRDFileProcessor
from .mpl_canvas import PlotSpectrum


class DefineAnyTemplate(PlotSpectrum):
    """Template to define integrals and backgrounds."""

    def __init__(self, crd: CRDFileProcessor, logy=True, theme: str = None) -> None:
        """Get a PyQt5 window to define the mass calibration for the given data.

        :param crd: The CRD file processor to work with.
        :param logy: Display the y axis logarithmically? Bottom set to 0.7
        :param theme: Theme to load, requires ``pyqtdarktheme`` to be installed
        """
        super().__init__(crd, logy=logy, theme=theme)

        # create a matpotlib canvas
        self.sc.mouse_right_press_position.connect(self.mouse_right_press)
        self.sc.mouse_right_release_position.connect(self.mouse_right_released)

        # buttons in bottom layout
        cancel_button = QtWidgets.QPushButton("Cancel")
        cancel_button.setToolTip("Cancel defining integrals.")
        cancel_button.clicked.connect(lambda: self.close())
        self.apply_button = QtWidgets.QPushButton("Apply")

        self.apply_button.setToolTip("Apply integrals.")
        self.apply_button.clicked.connect(lambda: self.apply())

        # set layout of bottom part
        self.bottom_layout.addStretch()
        self.bottom_layout.addWidget(cancel_button)
        self.bottom_layout.addWidget(self.apply_button)

        # some variables
        self._last_xpos = None

        # if integrals already exist
        if (crd_int := crd.def_integrals) is None:
            self.int_names = []
            self.int_values = []
        else:
            self.int_names, int_values = crd_int
            self.int_values = list(int_values)

        self.button_header = None
        self.button_tooltip = None

        # plot data
        self.plot_ms()

    def apply(self):
        """Apply the mass calibration and return it."""
        raise NotImplementedError

    def button_pressed(self, name: str):
        """Define action for left click of a peak button.

        :param name: Name of the peak.
        """  # noqa: DAR401
        raise NotImplementedError

    def check_peak_overlapping(self, peak_pos: np.array) -> Union[List, None]:
        """Check if a given peak position overlaps with any other peaks.

        :param peak_pos: Position of peak, 2 entry array with from, to.

        :return: List of all peak names that it overlaps or ``None``.
        """
        left = peak_pos[0]
        right = peak_pos[1]

        outlist = []
        for it, [dleft, dright] in enumerate(self.int_values):
            name = self.int_names[it]
            if (
                (dleft <= left < dright)
                or (dleft < right <= dright)
                or (left < dleft and right > dright)
            ):
                outlist.append(name)

        if outlist:
            return outlist
        else:
            return

    def clear_layout(self, layout) -> None:
        """Clear a given layout of all widgets, etc."""
        while layout.count():
            child = layout.takeAt(0)
            if child.widget() is not None:
                child.widget().deleteLater()
            elif child.layout() is not None:
                self.clear_layout(child.layout())

    def create_buttons(self):
        """Create the buttons in the right menubar."""
        self.clear_layout(self.right_layout)

        # add text to layout
        if self.button_header is not None:
            self.right_layout.addWidget(QtWidgets.QLabel(self.button_header))

        # create buttons with functions to delete values
        for name in self.int_names:
            button = QtWidgets.QPushButton(name)
            if self.button_tooltip is not None:
                button.setToolTip(self.button_tooltip)
            button.pressed.connect(lambda val=name: self.button_pressed(val))
            self.right_layout.addWidget(button)
        self.right_layout.addStretch()

    def mouse_right_press(self, xpos: float) -> None:
        """Act on right mouse button pressed.

        :param xpos: Position on x axis.
        """
        self._last_xpos = xpos

    def mouse_right_released(self, xpos: float) -> None:
        """Right mouse button was released.

        :param xpos: Position on x axis.
        """
        peak = np.sort(np.array([self._last_xpos, xpos], dtype=float))
        self.user_input(peak)

    def peaks_changed(self):
        """Go through the list of peaks, make buttons and shade areas."""
        raise NotImplementedError

    def shade_peaks(self):
        """Shade the peaks with given integrals."""
        # clear plot but keep axes limits (in case zoomed)
        xax_lims = self.axes.get_xlim()
        yax_lims = self.axes.get_ylim()
        self.axes.clear()
        self.plot_ms()
        self.axes.set_xlim(xax_lims)
        self.axes.set_ylim(yax_lims)

        # shade peaks
        for it, peak_pos in enumerate(self.int_values):
            indexes = np.where(
                np.logical_and(self.crd.mass > peak_pos[0], self.crd.mass < peak_pos[1])
            )

            self.axes.fill_between(
                self.crd.mass[indexes],
                self.crd.data[indexes],
                color=tableau_color(it),
                linewidth=0.3,
            )

    def sort_integrals(self):
        """Sort the names and integrals."""
        if len(self.int_names) > 1:
            sorted_zip = sorted(zip(self.int_names, self.int_values))
            self.int_names = [i for i, j in sorted_zip]
            self.int_values = [j for i, j in sorted_zip]

    def user_input(self, peak_pos: np.array, name: str = "") -> None:
        """Query user for position.

        :param peak_pos: Sorted array, left and right position of peak.
        :param name: Name to preset the line-edit with.
        """  # noqa: DAR401
        raise NotImplementedError


class DefineBackgrounds(DefineAnyTemplate):
    """QMainWindow to define backgrounds."""

    signal_backgrounds_defined = QtCore.pyqtSignal()

    def __init__(self, crd: CRDFileProcessor, logy=True, theme: str = None) -> None:
        """Get a PyQt5 window to define backgrounds for the given integrals.

        :param crd: The CRD file processor to work with.
        :param logy: Display the y axis logarithmically? Bottom set to 0.7
        :param theme: Theme to load, requires ``pyqtdarktheme`` to be installed
        """
        super().__init__(crd, logy=logy, theme=theme)

        self.setWindowTitle("Define integrals")

        # if backgrounds already exist
        if (crd_int := crd.def_backgrounds) is None:
            self.bg_names = []
            self.bg_values = []
        else:
            self.bg_names, bg_values = crd_int
            self.bg_values = list(bg_values)

        # temp variables
        self._selected_peak_name = None

        self.status_bar.showMessage(
            "Right click, drag, and release to define background."
        )

        self.button_header = "Click to delete:"
        self.button_tooltip = (
            "Click to delete all background definitions for this peak."
        )

        QtWidgets.QInputDialog()

        self.sort_integrals()
        self.create_buttons()
        self.shade_peaks()
        self.shade_backgrounds()

    def apply(self):
        """Apply the mass calibration and return it."""
        if self.bg_names:
            self.crd.def_backgrounds = self.bg_names, np.array(self.bg_values)
        else:
            self.crd.def_backgrounds = None
        self.signal_backgrounds_defined.emit()
        self.close()

    def button_pressed(self, name: str):
        """Delete a peak from consideration names and values list.

        :param name: Name of the peak.
        """
        comp = True
        while comp:
            try:
                index_to_pop = self.bg_names.index(name)
                self.bg_names.pop(index_to_pop)
                self.bg_values.pop(index_to_pop)
            except ValueError:
                comp = False

        self.peaks_changed()

    def peaks_changed(self):
        """Go through the list of peaks, make buttons and shade areas."""
        self.shade_peaks()  # also restarts the canvas
        self.shade_backgrounds()

    def shade_backgrounds(self):
        """Go through background list and shade them.

        .. note:: Canvas is not cleared prior to this!
        """
        for it, peak_pos in enumerate(self.bg_values):
            int_name_index = self.int_names.index(self.bg_names[it])
            col = tableau_color(int_name_index)

            self.axes.axvspan(
                peak_pos[0], peak_pos[1], linewidth=0, color=col, alpha=0.25
            )

        self.sc.draw()

    def user_input(self, bg_pos: np.array, name: str = "") -> None:
        """Query user for position of background.

        :param bg_pos: Sorted array, left and right position of background.
        :param name: Name of the peak, for reloading the UI.
        """
        # dialog
        def set_selected_peak_name(val):
            """Set the selected peak_name to the value provided by signal."""
            self._selected_peak_name = val

        dlg = PeakDialog(
            self, peak_names=self.int_names, desc="Select a peak for this background:"
        )
        dlg.peak_selected_signal.connect(set_selected_peak_name)

        if not dlg.exec():
            return
        else:
            name = self._selected_peak_name

        # process peak
        if ovl_names := self.check_peak_overlapping(bg_pos):
            ovl_names = np.array(ovl_names)

            # check if overlapping with more than one peak
            if np.where(ovl_names != name)[0].size > 0:
                question = QtWidgets.QMessageBox.question(
                    self,
                    "Overlap detected",
                    f"The selected background overlaps with a peak other than {name}. "
                    f"Is this what you wanted?",
                )
                if question == QtWidgets.QMessageBox.StandardButton.No:
                    return

            # peak itself is included in overlap, cut it out
            if np.where(ovl_names == name)[0].size > 0:
                bg_l, bg_r = bg_pos
                peak_l, peak_r = self.int_values[self.int_names.index(name)]

                # peak is inside background
                if bg_l < peak_l and bg_r > peak_r:
                    self.bg_names.append(name)
                    self.bg_values.append(np.array([bg_l, peak_l]))
                    self.bg_names.append(name)
                    self.bg_values.append(np.array([peak_r, bg_r]))
                    self.peaks_changed()
                    return
                elif bg_l < peak_l < bg_r:
                    bg_pos[1] = peak_l
                elif bg_l < peak_r < bg_r:
                    bg_pos[0] = peak_r

        self.bg_names.append(name)
        self.bg_values.append(bg_pos)
        self.peaks_changed()


class DefineIntegrals(DefineAnyTemplate):
    """QMainWindow to define integrals."""

    signal_integrals_defined = QtCore.pyqtSignal()

    def __init__(self, crd: CRDFileProcessor, logy=True, theme: str = None) -> None:
        """Get a PyQt5 window to define integrals in the given mass spectrum.

        :param crd: The CRD file processor to work with.
        :param logy: Display the y axis logarithmically? Bottom set to 0.7
        :param theme: Theme to load, requires ``pyqtdarktheme`` to be installed
        """
        super().__init__(crd, logy=logy, theme=theme)

        self.setWindowTitle("Define integrals")

        self.status_bar.showMessage(
            "Right click, drag, and release to define integral."
        )

        self.button_header = "Click to delete:"
        self.button_tooltip = "Click to delete this integral definition."

        self.sort_integrals()
        self.create_buttons()
        self.shade_peaks()

    def apply(self):
        """Apply the mass calibration and return it."""
        if self.int_names:
            self.crd.def_integrals = self.int_names, np.array(self.int_values)
        else:
            self.crd.def_integrals = None
        self.signal_integrals_defined.emit()
        self.close()

    def button_pressed(self, name: str):
        """Delete a peak from consideration names and values list.

        :param name: Name of the peak.
        """
        index_to_pop = self.int_names.index(name)
        self.int_names.pop(index_to_pop)
        self.int_values.pop(index_to_pop)
        self.peaks_changed()

    def peaks_changed(self):
        """Go through the list of peaks, make buttons and shade areas."""
        # sort the integrals
        self.sort_integrals()

        # buttons
        self.create_buttons()

        # shade
        self.shade_peaks()

    def user_input(self, peak_pos: np.array, name: str = "") -> None:
        """Query user for position.

        :param peak_pos: Sorted array, left and right position of peak.
        :param name: Name to pre-populate line-edit with.
        """
        # check for overlap:
        if self.check_peak_overlapping(peak_pos):
            QtWidgets.QMessageBox.warning(
                self,
                "Overlap detected",
                "Your peak overlaps with one or more other peaks. Try again.",
            )
            return

        user_in = QtWidgets.QInputDialog.getText(
            self,
            "Name",
            f"Please name the integral from {round(peak_pos[0], 2)} to "
            f"{round(peak_pos[1], 2)} amu.",
            text=name,
        )

        if user_in[1]:
            name = user_in[0]
            if name == "":
                QtWidgets.QMessageBox.warning(
                    self,
                    "Invalid name",
                    "Please enter a name for the integral or press Cancel.",
                )
            elif name in self.int_names:
                QtWidgets.QMessageBox.warning(
                    self,
                    "Name exists",
                    "Name already exists, please enter another name.",
                )
            else:
                self.int_names.append(name)
                self.int_values.append(peak_pos)
                self.peaks_changed()
                return
        else:
            return

        self.user_input(peak_pos, name)


class PeakDialog(QtWidgets.QDialog):
    """Custom QDialog to select from a list of buttons.

    If a button is pressed, a signal with its name as a string is emitted and the
    dialog is closed with an ``accept``. If the window is closed otherwise,
    a ``reject`` is sent.
    """

    peak_selected_signal = QtCore.pyqtSignal(str)

    def __init__(self, parent, peak_names=None, desc=None):
        """Initialize a peak dialog.

        .. note:: The parameters must be given as keyword argument.

        :param parent: Parent class for dialog.
        :param peak_names: Name of the peaks to be displayed. If None, will reject.
        :param desc: Description for user.
        """
        super().__init__(parent)

        if not peak_names:  # quit if no names were given
            self.reject()

        self.setWindowTitle("Pick from given list")

        # buttons
        button_layout = QtWidgets.QHBoxLayout()
        for it, name in enumerate(peak_names):
            button = QtWidgets.QPushButton(name)
            button.pressed.connect(lambda val=name: self.peak_selected(val))
            button_layout.addWidget(button)
            if it < len(peak_names) - 1:
                button_layout.addStretch()

        # main layout
        layout = QtWidgets.QVBoxLayout()

        # description
        if desc is not None:
            layout.addWidget(QtWidgets.QLabel(desc))
        layout.addLayout(button_layout)

        self.setLayout(layout)

    def peak_selected(self, val):
        """Emit signal with the peak name and exit out of dialog with accept.

        :param val: Name of the selected peak.
        """
        self.peak_selected_signal.emit(val)
        self.accept()


def define_backgrounds_app(crd: CRDFileProcessor, logy: bool = True) -> None:
    """Create a PyQt5 app for defining backgruonds.

    :param crd: CRD file to calibrate for.
    :param logy: Should the y axis be logarithmic? Defaults to True.
    """
    app = QtWidgets.QApplication(sys.argv)
    window = DefineBackgrounds(crd, logy=logy)
    window.show()
    app.exec()


def define_integrals_app(crd: CRDFileProcessor, logy: bool = True) -> None:
    """Create a PyQt5 app for defining integrals.

    :param crd: CRD file to calibrate for.
    :param logy: Should the y axis be logarithmic? Defaults to True.
    """
    app = QtWidgets.QApplication(sys.argv)
    window = DefineIntegrals(crd, logy=logy)
    window.show()
    app.exec()


def tableau_color(it: int = 0) -> str:
    """Return nth color from matplotlib TABLEAU_COLORS.

    If out of range, start at beginning.

    :param it: Which tableau color to get.

    :return: Matplotlib color string.
    """
    cols = list(mcolors.TABLEAU_COLORS.values())
    return cols[it % len(cols)]



"""Fixtures and configuration for pytest."""

from pathlib import Path
import struct

from typing import Tuple
import pytest
import numpy as np

import rimseval


# FIXTURES #


@pytest.fixture(scope="function")
def crd_file(tmpdir) -> Tuple[dict, np.ndarray, np.ndarray, Path]:
    """Create and write a CRD file to a temporary folder, then return it.

    :return: Header as dictionary, ions_per_shot, arrival_bins, Path to file
    """
    binLength = 100
    ions_per_shot = np.array([1, 0, 4, 1, 2])
    all_tofs = np.array([231, 471, 182, 221, 959, 448, 281, 221])
    # set header values for the file
    header = {
        "fileID": b"CRD\0",
        "startDateTime": b"2021:07:10 11:41:13\0",
        "minVer": 0,
        "majVer": 1,
        "sizeOfHeaders": 88,
        "shotPattern": 1,
        "tofFormat": 1,
        "polarity": 1,
        "binLength": binLength,
        "binStart": all_tofs.min(),
        "binEnd": all_tofs.max(),
        "xDim": 0,
        "yDim": 0,
        "shotsPerPixel": 0,
        "pixelPerScan": 0,
        "nofScans": 0,
        "nofShots": ions_per_shot.shape[0],
        "deltaT": 0.0,
    }
    # data format for packing
    fmt_hdr = (
        ("fileID", "4s"),
        ("startDateTime", "20s"),
        ("minVer", "<H"),
        ("majVer", "<H"),
        ("sizeOfHeaders", "<I"),
        ("shotPattern", "<I"),
        ("tofFormat", "<I"),
        ("polarity", "<I"),
        ("binLength", "<I"),
        ("binStart", "<I"),
        ("binEnd", "<I"),
        ("xDim", "<I"),
        ("yDim", "<I"),
        ("shotsPerPixel", "<I"),
        ("pixelPerScan", "<I"),
        ("nofScans", "<I"),
        ("nofShots", "<I"),
        ("deltaT", "<d"),
    )
    # create crd content
    crd_cont = b""
    for hdr_name, fmt in fmt_hdr:  # write header
        crd_cont += struct.pack(fmt, header[hdr_name])
    bin_index = 0
    for ion in ions_per_shot:  # write data
        crd_cont += struct.pack("<I", ion)
        for it in range(ion):
            crd_cont += struct.pack("<I", all_tofs[bin_index])
            bin_index += 1
    crd_cont += struct.pack("4s", b"OK!\0")
    # write the CRD file
    fname = tmpdir.join("testfile.crd")
    fname.write_binary(crd_cont)
    return header, ions_per_shot, all_tofs, fname


@pytest.fixture(scope="function")
def crd_proc_mock(mocker):
    """Create a mocker instance for the CRDFileProcessor class and return it."""
    crd_mock = mocker.MagicMock()
    crd_mock.__class__ = rimseval.processor.CRDFileProcessor
    return crd_mock


@pytest.fixture(scope="function")
def init_lst_proc():
    """Clean initialization of class and reset to defaults after usage."""
    # spin up class
    cls = rimseval.data_io.LST2CRD()

    yield cls

    # set defaults back that are specified in initialization
    cls._file_name = None
    cls._channel_data = None
    cls._channel_tag = None

    # reset all other variables
    cls._file_info = {}  # dictionary with parsed header info
    cls._data_format = None
    cls._data_signal = None  # main signal data
    cls._tags = None  # tag data



"""PyTest fixtures for unit tests in data_io."""

import os

from pathlib import Path
import pytest


@pytest.fixture
def crd_data(request) -> Path:
    """Provides the path to the `crd_data` folder.

    This folder contains simple, small CRD files with real data for testing.

    :return: Path to the folder
    """
    curr = Path(request.fspath).parents[0]
    return Path(curr).joinpath("../crd_data").absolute()


@pytest.fixture
def lst_crd_path(request):
    """Provides the path to the `lst_crd_files` folder.

    :return: Path to the folder
    :rtype: Path
    """
    curr = Path(request.fspath).parents[0]
    return Path(curr).joinpath("lst_crd_files").absolute()



"""Tests for export routines."""

from hypothesis import given, strategies as st
import numpy as np
from pathlib import Path
import pytest

from rimseval.processor import CRDFileProcessor
from rimseval.data_io import export


def test_bin_array_avg():
    """Bin an array by averaging."""
    arr_in = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

    bins = 2
    arr_exp = np.array([1.5, 3.5, 5.5, 7.5, 9.5])
    arr_rec = export._bin_array_avg(arr_in, bins)
    np.testing.assert_equal(arr_exp, arr_rec)

    bins = 3
    arr_exp = np.array([2, 5, 8])
    arr_rec = export._bin_array_avg(arr_in, bins)
    np.testing.assert_equal(arr_exp, arr_rec)


def test_bin_array_sum():
    """Bin an array by summing."""
    arr_in = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10])

    bin = 2
    arr_exp = np.array([3, 7, 11, 15, 19])
    arr_rec = export._bin_array_sum(arr_in, bin)
    np.testing.assert_equal(arr_exp, arr_rec)

    bin = 3
    arr_exp = np.array([6, 15, 24])
    arr_rec = export._bin_array_sum(arr_in, bin)
    np.testing.assert_equal(arr_exp, arr_rec)


def test_tof_spectrum(tmpdir, crd_file):
    """Export time of flight spectrum."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()

    file_out = Path(tmpdir.strpath).joinpath("tof")
    export.tof_spectrum(crd, file_out, bins=2)
    assert file_out.with_suffix(".csv").is_file()


def test_mass_spectrum(tmpdir, crd_file):
    """Export time of flight and mass spectrum."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))
    crd.def_mcal = np.array([[1, 2], [10, 20]])
    crd.spectrum_full()
    crd.mass_calibration()

    file_out = Path(tmpdir.strpath).joinpath("mass")
    export.mass_spectrum(crd, file_out, bins=2)
    assert file_out.with_suffix(".csv").is_file()



"""Test for Peirce criterion data rejection."""


from hypothesis import given, strategies as st
import pytest
import numpy as np

from rimseval.utilities import peirce


def test_peirce_criterion_ldiv_zero():
    """Choose a minimum ldiv if it is zero (erfc turns 0 at large values)."""
    # define values such that it turns out to be zero
    n_tot = 100
    n = 1e3  # this brings `ldiv` numerically to zero
    m = 1

    _ = peirce.peirce_criterion(n_tot, n, m)


@given(n_hyp=st.integers(), m_hyp=st.integers())
def test_peirce_criterion_n_tot_zero(n_hyp, m_hyp):
    """Check that setting number of observations to zero returns zero always."""
    n_tot = 0
    expected = 0
    received = peirce.peirce_criterion(n_tot, n_hyp, m_hyp)
    assert expected == received


def test_peirce_criterion_x2_less_zero():
    """Ensure that in while-loop, x2 is set to zero if it is smaller than zero."""
    # define values that get x2 < 0
    n_tot = 2
    n = 1
    m = 5

    expected = 0
    received = peirce.peirce_criterion(n_tot, n, m)

    assert expected == received


def test_peirce_ross_2003_example():
    """Run example in Ross (2003) through Peirce criterion."""
    data = np.array([102.2, 90, 99, 102, 103, 100.2, 89, 98.1, 101.5, 102])
    outliers_exp = np.array([89, 90])
    outliers_ind = np.array([1, 6])  # index of outliers in original array
    avg_exp = 100.9
    std_exp = 1.66

    avg_rec, std_rec, outliers_rec, outliers_ind_rec = peirce.reject_outliers(data)

    assert avg_rec == pytest.approx(avg_exp, abs=0.1)
    assert std_rec == pytest.approx(std_exp, abs=0.1)
    np.testing.assert_equal(outliers_rec, outliers_exp)
    np.testing.assert_equal(outliers_ind_rec, outliers_ind)



"""Test for fitting utilities."""

from hypothesis import given, strategies as st
import pytest
import numpy as np

from rimseval.utilities import fitting

# Hypothesis variables
my_float = st.floats(allow_nan=False)
my_tuple = st.tuples(my_float, my_float)


@given(xdat=st.lists(my_float), coeffs=st.tuples(my_float, my_float, my_float))
def test_gaussian(xdat, coeffs):
    """Return a Gaussian for given coefficients."""
    xdat = np.array(xdat)
    coeffs = np.array(coeffs)
    exp_value = coeffs[2] * np.exp(-(((xdat - coeffs[0]) / coeffs[1]) ** 2))
    np.testing.assert_equal(fitting.gaussian(xdat, coeffs), exp_value)


@given(
    xydat=st.lists(my_tuple, min_size=1),
    coeffs=st.tuples(my_float, my_float, my_float),
)
def test_residual_gaussian(xydat, coeffs):
    """Return a Gaussian for given coefficients."""
    xydat = np.array(xydat)
    xdat = xydat[:, 0]
    ydat = xydat[:, 1]

    coeffs = np.array(coeffs)

    exp_value = ydat - fitting.gaussian(xdat, coeffs)
    np.testing.assert_equal(fitting.residuals_gaussian(coeffs, ydat, xdat), exp_value)



"""Resonance Ionization Mass Spectrometry (RIMS) Data Evaluation for CRD Files."""

import iniabu

from . import data_io
from . import guis
from . import interfacer
from . import utilities
from .multi_proc import MultiFileProcessor
from .processor import CRDFileProcessor

ini = iniabu.IniAbu(database="nist")

__all__ = [
    "ini",
    "CRDFileProcessor",
    "data_io",
    "guis",
    "interfacer",
    "MultiFileProcessor",
    "utilities",
]

# Package information
__version__ = "2.0.0"

__title__ = "rimseval"
__description__ = (
    "Evaluate resonance ionization mass spectrometry measurements, correct, and "
    "filter. Contains a reader for crd files (Chicago Raw Data) but can also convert "
    "FastComTec lst files to crd."
)

__uri__ = "https://rimseval.readthedocs.io"
__author__ = "Reto Trappitsch"

__license__ = "MIT"
__copyright__ = "Copyright (c) 2022, Reto Trappitsch"



"""Interfacing functions to talk to settings, calibrations, GUIs, etc."""

import json
from pathlib import Path
from typing import Any

import numpy as np

from rimseval.compatibility.lion_eval import LIONEvalCal
from rimseval.processor import CRDFileProcessor


def read_lion_eval_calfile(crd: CRDFileProcessor, fname: Path = None) -> None:
    """Read a LIONEval calibration file and set it to instance of crd.

    LIONEval is the first, Python2.7 version of the data evaluation software. This
    routine takes an old calibration file if requested by the user and sets the
    mass calibration, integrals, and background correction information if present.

    :param crd: Instance of the CRDFileProcessor, since we need to set properties
    :param fname: Filename to mass calibration file. If `None`, try the same file name
        as for the CRD file, but with `.cal` as an extension.

    :raises OSError: Calibration file does not exist.
    """
    if fname is None:
        fname = crd.fname.with_suffix(".cal")

    if not fname.exists():
        raise OSError(f"The requested calibration file {fname} does not exist.")

    cal = LIONEvalCal(fname)

    if cal.mass_cal is not None:
        crd.def_mcal = cal.mass_cal

    names_int = None
    if cal.integrals:
        names_int = []
        areas_int = np.zeros((len(cal.integrals), 2))
        for it, line in enumerate(cal.integrals):
            names_int.append(line[0])
            areas_int[it][0] = line[1] - line[2]
            areas_int[it][1] = line[1] + line[3]
        crd.def_integrals = (names_int, areas_int)

    if cal.bg_corr and cal.integrals:  # w/o integrals, don't load bgs!
        names_bg = []
        areas_bg = []
        for line in cal.bg_corr:
            name = line[0]
            if name in names_int:  # must be the case for new program
                names_bg.append(name)
                areas_bg.append([line[2], line[3]])
                names_bg.append(name)
                areas_bg.append([line[4], line[5]])

        areas_bg = np.array(areas_bg)
        crd.def_backgrounds = (names_bg, areas_bg)

    if cal.applied_filters is not None:
        crd.applied_filters = cal.applied_filters


def load_cal_file(crd: CRDFileProcessor, fname: Path = None) -> None:
    """Load a calibration file from a specific path / name.

    :param crd: CRD Processor class to load into
    :param fname: Filename and path. If `None`, try file with same name as CRD file but
        `.json` suffix.

    :raises OSError: Calibration file does not exist.
    :raises OSError: JSON file cannot be decoded. JSON error message is returned too.
    """
    if fname is None:
        fname = crd.fname.with_suffix(".json")

    if not fname.exists():
        raise OSError(f"The requested calibration file {fname} does not exist.")

    with fname.open("r", encoding="utf-8") as fin:
        try:
            json_object = json.load(fin)
        except json.decoder.JSONDecodeError as orig_err:
            raise OSError(
                f"Cannot open the calibration file {fname.name}. JSON decode error."
            ) from orig_err

    def entry_loader(key: str, json_obj: Any) -> Any:
        """Return the value of a json_object dictionary if existent, otherwise None."""
        if key in json_obj.keys():
            return json_obj[key]
        else:
            return None

    # mass cal
    mcal = entry_loader("mcal", json_object)
    if mcal is not None:
        crd.def_mcal = np.array(mcal)

    # integrals
    names_int = entry_loader("integral_names", json_object)
    integrals = np.array(entry_loader("integrals", json_object))

    if names_int is not None and integrals is not None:
        crd.def_integrals = names_int, integrals

    # backgrounds
    names_bgs = entry_loader("background_names", json_object)
    backgrounds = np.array(entry_loader("backgrounds", json_object))

    if names_bgs is not None and backgrounds is not None:
        crd.def_backgrounds = names_bgs, backgrounds

    # applied filters
    applied_filters = entry_loader("applied_filters", json_object)
    if applied_filters is not None:
        crd.applied_filters = applied_filters


def save_cal_file(crd: CRDFileProcessor, fname: Path = None) -> None:
    """Save a calibration file to a specific path / name.

    Note: The new calibration files are `.json` files and not `.cal` files.

    :param crd: CRD class instance to read all the data from.
    :param fname: Filename to save to to. If None, will save in folder / name of
        original crd file name, but with '.cal' ending.
    """
    if fname is None:
        fname = crd.fname.with_suffix(".json")

    cal_to_write = {}

    # mass cal
    if crd.def_mcal is not None:
        cal_to_write["mcal"] = crd.def_mcal.tolist()

    # integrals
    if crd.def_integrals is not None:
        names_int, integrals = crd.def_integrals
        cal_to_write["integral_names"] = names_int
        cal_to_write["integrals"] = integrals.tolist()

    # backgrounds
    if crd.def_backgrounds is not None:
        names_bg, backgrounds = crd.def_backgrounds
        cal_to_write["background_names"] = names_bg
        cal_to_write["backgrounds"] = backgrounds.tolist()

    # filters
    if crd.applied_filters != {}:
        cal_to_write["applied_filters"] = crd.applied_filters

    json_object = json.dumps(cal_to_write, indent=4)

    fname.write_text(json_object, encoding="utf-8")



"""This file contains utilities for processing crd files."""

from enum import Enum
import struct

from numba import njit
import numpy as np

# current default CRD header information, packed and ready to be written
CURRENT_DEFAULTS = {
    "fileID": struct.pack("4s", bytes("CRD", "utf-8")),
    "minVer": struct.pack("<H", 0),
    "majVer": struct.pack("<H", 1),
    "sizeOfHeaders": struct.pack("<I", 88),
    "shotPattern": struct.pack("<I", 0),
    "tofFormat": struct.pack("<I", 1),
    "polarity": struct.pack("<I", 1),
    "xDim": struct.pack("<I", 0),
    "yDim": struct.pack("<I", 0),
    "shotsPerPixel": struct.pack("<I", 0),
    "pixelPerScan": struct.pack("<I", 0),
    "nOfScans": struct.pack("<I", 0),
    "deltaT": struct.pack("<d", 0),
    "eof": struct.pack("4s", bytes("OK!", "utf-8")),
}

HEADER_START = (
    ("fileID", 4, "4s"),
    ("startDateTime", 20, "20s"),
    ("minVer", 2, "<H"),
    ("majVer", 2, "<H"),
    ("sizeOfHeaders", 4, "<I"),
)


class CRDHeader(Enum):
    """Enum class for CRD header.

    The start must always be the same as HEADER_START above, however, the other fields
    might vary depending on the header that is being used. The header is called by its
    version number. Note that the letter `v` precedes the number and that the period
    is replaced with the letter `p`.

    Format is always: Name, length, struct unpack qualifier
    """

    v1p0 = (
        ("shotPattern", 4, "<I"),
        ("tofFormat", 4, "<I"),
        ("polarity", 4, "<I"),
        ("binLength", 4, "<I"),
        ("binStart", 4, "<I"),
        ("binEnd", 4, "<I"),
        ("xDim", 4, "<I"),
        ("yDim", 4, "<I"),
        ("shotsPerPixel", 4, "<I"),
        ("pixelPerScan", 4, "<I"),
        ("nofScans", 4, "<I"),
        ("nofShots", 4, "<I"),
        ("deltaT", 8, "<d"),
    )


@njit
def shot_to_tof_mapper(ions_per_shot: np.array) -> np.array:  # pragma: nocover
    """Mapper for ions_to_shot to all_tofs.

    Takes ions_per_shot array and creates a mapper that describes which ranges in the
    all_tofs array a given shot refers to.

    :param ions_per_shot: Ion per shots array.

    :return: Mappeing array shots to ToF.
    """
    mapper = np.zeros((ions_per_shot.shape[0], 2), dtype=np.int32)
    curr_ind = 0
    for it, ion_per_shot in enumerate(ions_per_shot):
        mapper[it][0] = curr_ind
        mapper[it][1] = curr_ind + ion_per_shot
        curr_ind += ion_per_shot
    return mapper



"""Function test for processor class properties."""

from pathlib import Path

import hypothesis
from hypothesis import given, HealthCheck, settings, strategies as st
import pytest
import numpy as np

from rimseval.processor import CRDFileProcessor


def test_def_backgrounds(crd_file):
    """Get / set background definitions."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))

    # backgrounds not defined at start
    assert crd.def_backgrounds is None

    peak_names = ["54Fe", "54Fe"]
    peak_limits = np.array([[53.4, 53.6], [54.4, 54.6]])

    # set some backgrounds
    crd.def_backgrounds = (peak_names, peak_limits)
    ret_names, ret_limts = crd.def_backgrounds
    assert ret_names == peak_names
    np.testing.assert_equal(ret_limts, peak_limits)

    # unset backgrounds
    crd.def_backgrounds = None
    assert crd.def_backgrounds is None


def test_def_backgrounds_value_error(crd_file):
    """Raise value errors if backgrounds defined with wrong shape."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))

    with pytest.raises(ValueError):
        crd.def_backgrounds = ["54Fe", 53, 54]
    with pytest.raises(ValueError):
        crd.def_backgrounds = (["54Fe"], np.array([]))
    with pytest.raises(ValueError):
        crd.def_backgrounds = (["54Fe"], np.array([[1, 2, 3]]))


def test_def_mcal(crd_file):
    """Get / set mass calibration definitions."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))

    assert crd.def_mcal is None

    mass_cal = np.array([[64, 223.0], [66, 337.0]])
    crd.def_mcal = mass_cal
    np.testing.assert_equal(mass_cal, crd.def_mcal)


def test_def_mcal_type_error(crd_file):
    """Raise TypeError if mass calibration is set with none numpy array."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))

    mass_cal = [[64, 223.0], [66, 337.0]]

    with pytest.raises(TypeError):
        crd.def_mcal = mass_cal


def test_def_mcal_value_error(crd_file):
    """Raise ValueError if mass_cal is of wrong shape."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))

    with pytest.raises(ValueError):
        crd.def_mcal = np.array([[1.0, 2.0]])
    with pytest.raises(ValueError):
        crd.def_mcal = np.array([[1.0], [2.0], [3.0]])


def test_def_integrals(crd_file):
    """Get / set integral definitions."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))

    # integrals not defined at start
    assert crd.def_integrals is None

    peak_names = ["54Fe", "64Ni"]
    peak_limits = np.array([[53.8, 54.2], [63.5, 64.5]])

    # set some integrals
    crd.def_integrals = (peak_names, peak_limits)
    ret_names, ret_limts = crd.def_integrals
    assert ret_names == peak_names
    np.testing.assert_equal(ret_limts, peak_limits)

    # unset integrals
    crd.def_integrals = None
    assert crd.def_integrals is None


def test_def_integrals_value_error(crd_file):
    """Raise value errors if integrals defined with wrong shape or non-unique peaks."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))

    with pytest.raises(ValueError):
        crd.def_integrals = ["54Fe", 53, 54]
    with pytest.raises(ValueError):
        crd.def_integrals = (["54Fe"], np.array([]))
    with pytest.raises(ValueError):
        crd.def_integrals = (["54Fe"], np.array([[1, 2, 3]]))

    with pytest.raises(ValueError):
        crd.def_integrals = (["54Fe", "54Fe"], np.array([[53.8, 54.2], [63.5, 64.5]]))


@given(value=st.floats(allow_nan=False))
@settings(suppress_health_check=[HealthCheck.function_scoped_fixture], deadline=500)
def test_peak_fwhm(crd_file, value):
    """Get / set peak full width at half max."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))

    crd.peak_fwhm = value
    assert crd.peak_fwhm == value


@given(value=st.floats(allow_nan=False))
@settings(suppress_health_check=[HealthCheck.function_scoped_fixture], deadline=500)
def test_us_to_chan(crd_file, value):
    """Get / set conversion factor for microseconds to channel."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))

    crd.us_to_chan = value
    assert crd.us_to_chan == value



"""Interactive mass calibration using matplotlib's qtagg backend."""

from functools import partial
import sys
from typing import List, Tuple, Union

import numpy as np
from PyQt6 import QtCore, QtWidgets


from rimseval.processor import CRDFileProcessor
import rimseval.processor_utils
from rimseval.processor_utils import gaussian_fit_get_max
from rimseval.utilities import ini
from .mpl_canvas import PlotSpectrum


class CreateMassCalibration(PlotSpectrum):
    """QMainWindow to create a mass calibration."""

    signal_calibration_applied = QtCore.pyqtSignal()

    def __init__(
        self, crd: CRDFileProcessor, logy=True, mcal: np.array = None, theme=None
    ) -> None:
        """Get a PyQt6 window to define the mass calibration for the given data.

        :param crd: The CRD file processor to work with.
        :param logy: Display the y axis logarithmically? Bottom set to 0.7
        :param mcal: Existing mass calibration.
        :param theme: Theme to load, requires ``pyqtdarktheme`` to be installed
        """
        super().__init__(crd, logy=logy, theme=theme)
        self.setWindowTitle("Create mass calibration")
        self.theme = theme

        # create a matpotlib canvas
        self.sc.mouse_right_press_position.connect(self.right_click_event)

        # buttons in bottom layout
        self.undo_button = QtWidgets.QPushButton("Undo")
        self.undo_button.setToolTip("Undo last peak calibration, see status bar.")
        self.undo_button.clicked.connect(lambda: self.undo_last_mcal())

        cancel_button = QtWidgets.QPushButton("Cancel")
        cancel_button.setToolTip("Cancel mass calibration.")
        cancel_button.clicked.connect(lambda: self.close())
        self.apply_button = QtWidgets.QPushButton("Apply")

        self.apply_button.setToolTip("Apply mass calibration.")
        self.apply_button.clicked.connect(lambda: self.apply())

        # set layout of bottom part
        self.bottom_layout.addStretch()
        self.bottom_layout.addWidget(self.undo_button)
        self.bottom_layout.addWidget(cancel_button)
        self.bottom_layout.addWidget(self.apply_button)

        # some variables for guessing
        self._last_element = None
        self._mass = None  # mass array for guessing
        self._mass_axis = None

        # init mass calibration
        if mcal is None:
            self._mcal = []
        else:
            self._mcal = mcal.tolist()
        self.check_mcal_length()

        # help in statusbar
        self.status_bar.showMessage(
            "Please right-click on a peak to begin mass calibration."
        )

        # plot data
        self.plot_tof()

    def append_to_mcal(self, tof: float, mass: float) -> None:
        """Append a given value to the mass calibration.

        :param tof: Time of flight.
        :param mass: Mass.
        """
        self._mcal.append([tof, mass])
        self.check_mcal_length()

    def apply(self):
        """Apply the mass calibration and return it."""
        self.crd.def_mcal = np.array(self._mcal)
        self.crd.mass_calibration()
        self.signal_calibration_applied.emit()
        self.close()

    def check_mcal_length(self):
        """Check length of mcal to set button statuses, start guessing."""
        # apply button
        if len(self._mcal) >= 2:
            # button
            self.apply_button.setDisabled(False)
            # mass calibration
            self._mass, params = rimseval.processor_utils.mass_calibration(
                np.array(self._mcal), self.crd.tof, return_params=True
            )
            # plot secondary axis
            self.secondary_axis(params=params, visible=True)
        else:
            self.apply_button.setDisabled(True)
            self._mass = None
            self._last_element = None
            self.secondary_axis(visible=False)

        if len(self._mcal) > 0:
            self.undo_button.setDisabled(False)
        else:
            self.undo_button.setDisabled(True)

    def guess_mass(self, tof: float) -> str:
        """Guess the mass depending on what is currently set.

        If not drawn from iniabu, I assume the user wants even masses!

        :param tof: Time of Flight (us)

        :return: Guessed mass.
        """
        ind_tof = np.argmin(np.abs(self.crd.tof - tof))
        mass = self._mass[ind_tof]

        if self._last_element is not None:  # guess with iniabu
            name_iso, mass_iso = find_closest_iso(mass, self._last_element)
            if np.abs(mass_iso - mass) < 0.5:  # within half a mass
                return name_iso
            else:
                return find_closest_iso(mass)[0]
        else:  # guess mass from values
            return str(int(np.round(mass, 0)))

    def query_mass(self, tof: float) -> Union[float, None]:
        """Query mass from user.

        Query a mass from the user using a QInputDialog.

        :param tof: Time of flight of the clicked area.

        :return: Mass of the peak as given by user.
        """
        if self._mass is not None:
            guess = self.guess_mass(tof)
        else:
            guess = ""

        user_input = QtWidgets.QInputDialog.getText(
            self,
            "Calibrate Mass",
            f"Enter isotope name or mass for {tof:.2f}us.",
            text=guess,
        )

        def err_invalid_entry():
            """Show Error Message for invalid entry."""
            QtWidgets.QMessageBox.warning(
                self,
                "Invalid Input",
                "No valid input. Please enter a mass (number) or an isotope "
                "in the format, e.g., 46Ti, Ti46, or Ti-46.",
            )

        def err_invalid_isotope(iso):
            """Show Error Message for invalid isotope."""
            QtWidgets.QMessageBox.warning(
                self,
                "Invalid Isotope",
                f"Could not find an isotope with the name {iso}. "
                f"Please make sure you entered a valid isotope name, "
                f"e.g., 46Ti or Ti46.\n"
                f"It could also be that the isotope you requested is "
                f"currently not available in the database. In that case, "
                f"please enter the mass manually.",
            )

        if user_input[1]:
            if (iso := user_input[0]) == "":
                err_invalid_entry()
                return self.query_mass(tof)
            try:  # user input is a mass
                mass = float(iso)
            except ValueError:
                try:
                    mass = ini.iso[iso].mass
                    self._last_element = ini.iso[iso].name.split("-")[0]
                except IndexError:
                    err_invalid_isotope(iso)
                    return self.query_mass(tof)
            return mass
        else:
            return

    def right_click_event(self, xpos: float, *args, **kwargs) -> None:
        """Act on an emitted right click event."""
        mass = self.query_mass(xpos)
        if mass is None:
            return None  # user hit cancel

        min_value = xpos - 2 * self.crd.peak_fwhm
        max_value = xpos + 2 * self.crd.peak_fwhm
        window = np.where(
            np.logical_and(self.crd.tof > min_value, self.crd.tof < max_value)
        )

        xdata = self.crd.tof[window]
        ydata = self.crd.data[window]

        tof_max = gaussian_fit_get_max(xdata, ydata)

        # make sure max is in between min and max
        if tof_max <= xdata.min() or tof_max >= xdata.max():
            QtWidgets.QMessageBox.warning(
                self,
                "No peak found",
                "Couldn't find a peak. Please try again or choose a bigger peak.",
            )
            return None

        self.append_to_mcal(tof_max, mass)

        self.status_bar.showMessage(
            f"Peak with mass {mass:.2f} found at {tof_max:.2f}us."
        )

    def secondary_axis(self, params: List = None, visible: bool = False) -> None:
        """Toggle secondary axis.

        :param params: Parameters for the transfer functions t0 and const. Only
            required when visible=True.
        :param visible: Turn it on? If so, params are required.
        """
        if visible:
            if params is None:
                return
            else:
                if self._mass_axis is not None:
                    self._mass_axis.set_visible(False)
                self._mass_axis = self.axes.secondary_xaxis(
                    "top",
                    functions=(
                        partial(
                            rimseval.processor_utils.tof_to_mass,
                            tm0=params[0],
                            const=params[1],
                        ),
                        partial(
                            rimseval.processor_utils.mass_to_tof,
                            tm0=params[0],
                            const=params[1],
                        ),
                    ),
                )
                self._mass_axis.set_xlabel("Mass (amu)")
                if self.theme == "dark":
                    self._mass_axis.set_color("tab:orange")
                else:
                    self._mass_axis.set_color("tab:red")
                self.sc.draw()
        elif self._mass_axis is not None:
            self._mass_axis.set_visible(False)
            self._mass_axis = None
            self.sc.draw()

    def undo_last_mcal(self):
        """Undo the last mass calibration by popping the last entry of list."""
        tof, mass = self._mcal.pop()
        self.status_bar.showMessage(
            f"Deleted calibration with mass {mass:.2f} at {tof:.2f}us."
        )
        self.check_mcal_length()


def create_mass_cal_app(crd: CRDFileProcessor, logy: bool = True, theme=None) -> None:
    """Create a PyQt5 app for the mass cal window.

    :param crd: CRD file to calibrate for.
    :param logy: Should the y axis be logarithmic? Defaults to True.
    :param theme: Theme of GUI, requires ``pyqtdarktheme`` to be installed
    """
    app = QtWidgets.QApplication(sys.argv)
    mcal = crd.def_mcal
    window = CreateMassCalibration(crd, mcal=mcal, logy=logy, theme=theme)
    window.show()
    app.exec()


def find_closest_iso(mass: float, key: List = None) -> Tuple[str, float]:
    """Find closest iniabu isotope to given mass and return its name.

    If a key is given, will only consider that element, otherwise all.

    :param mass: Mass of the isotope to look for.
    :param key: An element or isotope key that is valid for iniabu.

    :return: Closest isotpoe, name and mass as tuple.
    """
    if key is None:
        key = list(ini.ele_dict.keys())
    isos = ini.iso[key]
    index = np.argmin(np.abs(isos.mass - mass))
    return isos.name[index], isos.mass[index]



"""Function tests for the multi file processor class."""

from pathlib import Path

import rimseval
from rimseval.multi_proc import MultiFileProcessor as mfp
from rimseval.processor import CRDFileProcessor


def test_mfp_num_of_files():
    """Return the number of files given to the multi processor."""
    files = [Path("a.crd"), Path("b.crd"), Path("c.crd")]

    crds = mfp(files)
    assert crds.num_of_files == len(files)


def test_mfp_peak_fwhm(crd_file):
    """Get / set the peak FWHM of the multip processor."""
    val = 42.0
    _, _, _, fname = crd_file
    files = [Path(fname)]

    crds = mfp(files)
    crds.peak_fwhm = val
    assert crds.peak_fwhm == val


def test_mfp_close_files(crd_file):
    """Close and delete a crd file, and associate it with None."""
    _, _, _, fname = crd_file
    files = [Path(fname)]

    crds = mfp(files)
    crds.open_files()

    crds.close_files()
    assert crds._files is None


def test_mfp_close_selected_files(crd_file):
    """Close 2 out of three selected files."""
    _, _, _, fname = crd_file
    files = [Path(fname)] * 5

    crds = mfp(files)
    crds.open_files()

    main_id = 3  # the main file

    ids_to_del = [0, 2]
    new_main_id = crds.close_selected_files(ids_to_del, main_id=main_id)
    assert new_main_id == 1  # main id update
    assert len(crds.files) == len(files) - len(ids_to_del)

    # delete main file
    assert crds.close_selected_files([new_main_id], main_id=new_main_id) == 0

    # delete one more without a main id file -> returns zero
    assert crds.close_selected_files([0]) == 0
    assert len(crds.files) == len(files) - len(ids_to_del) - 2


def test_mfp_load_calibration_pass_through(crd_file):
    """Do nothing if no calibration file is available."""
    _, _, _, fname = crd_file
    crds = mfp([Path(fname)])
    crds.load_calibrations()
    assert crds.files[0].def_mcal is None


def test_mfp_open_crd_file(crd_file):
    """Open a single file in the multifileprocessor."""
    _, _, _, fname = crd_file
    files = [Path(fname)]

    crds = mfp(files)
    crds.open_files()
    assert crds.files is not None
    for crd in crds.files:
        assert isinstance(crd, CRDFileProcessor)


def test_mfp_open_additional_file(crd_file):
    """Open additional file in MFP with single file already opened."""
    _, _, _, fname = crd_file

    crds = mfp([Path(fname)])
    crds.read_files()

    assert len(crds.files) == 1

    crds.open_additional_files([Path(fname), Path(fname)], read_files=True)
    assert len(crds.files) == 3

    # ensure all files were read
    for crd in crds.files:
        assert crd.tof is not None


def test_mfp_read_crd_file(crd_file):
    """Read a single crd file without opening it (opening happens automatically)."""
    _, _, _, fname = crd_file
    files = [Path(fname)]

    crds = mfp(files)
    crds.read_files()

    for crd in crds.files:
        assert crd.tof is not None



"""Tests for processors utilities."""

import numpy as np

import rimseval.data_io.crd_utils as cu


def test_shot_to_tof_mapper():
    """Map ions per shot to all_tofs array."""
    ions_per_shot = np.array([0, 0, 3, 5, 0, 7])
    mapper_exp = np.array([[0, 0], [0, 0], [0, 3], [3, 8], [8, 8], [8, 15]])
    mapper_rec = cu.shot_to_tof_mapper(ions_per_shot)
    np.testing.assert_equal(mapper_rec, mapper_exp)



"""Test for Peirce criterion data rejection."""


import pytest
import numpy as np

from rimseval.utilities import utils


def test_not_index():
    """Return all indices not of length that are not in a given array."""
    # docstring example
    a = np.arange(7)
    b = np.where(a < 4)[0]
    expected = np.array([4, 5, 6])
    np.testing.assert_equal(expected, utils.not_index(b, len(a)))

    # filter a bit further
    b = np.where(np.logical_or(a < 4, a > 5))[0]
    expected = np.array([4, 5])
    np.testing.assert_equal(expected, utils.not_index(b, len(a)))

    # only one entry left
    b = np.where(a < 1)[0]
    expected = np.array([1, 2, 3, 4, 5, 6])
    np.testing.assert_equal(expected, utils.not_index(b, len(a)))


def test_not_index_empty_ind():
    """Return all indices if `ind` is empty."""
    a = np.arange(7)
    b = np.where(a < 0)[0]
    expected = np.arange(len(a))
    np.testing.assert_equal(expected, utils.not_index(b, len(a)))


def test_not_index_value_error():
    """Raise a value error if max index is larger than length."""
    ind = np.arange(10)
    with pytest.raises(ValueError):
        utils.not_index(ind, 5)



"""PyTest fixtures for unit tests."""

from pathlib import Path
import pytest


@pytest.fixture
def crd_data(request) -> Path:
    """Provides the path to the `crd_data` folder.

    This folder contains simple, small CRD files with real data for testing.

    :return: Path to the folder
    """
    curr = Path(request.fspath).parents[0]
    return Path(curr).joinpath("crd_data").absolute()


@pytest.fixture
def legacy_files_path(request) -> Path:
    """Provides the path to the `legacy_files` folder.

    :return: Path to the folder
    """
    curr = Path(request.fspath).parents[0]
    return Path(curr).joinpath("legacy_files").absolute()


@pytest.fixture
def macro_path(request) -> Path:
    """Provides the path to this folder where the macros are.

    :return: Path to the folder
    """
    return Path(request.fspath).parents[0].absolute().joinpath("macro/")



"""Tests for `test_lst_to_crd.py`"""

from hypothesis import given, strategies as st
import numpy as np
from pathlib import Path
import pytest

import rimseval


# PROPERTIES #


def test_channel_data(init_lst_proc):
    """Get / set the number of the channel the data is in."""
    channel_number = 42
    init_lst_proc.channel_data = channel_number
    assert init_lst_proc.channel_data == channel_number


def test_channel_data_wrong_type(init_lst_proc):
    """Raise a type error when a wrong type is set for the channel number."""
    wrong_type = "42"
    with pytest.raises(TypeError) as exc_info:
        init_lst_proc.channel_data = wrong_type
    exc_msg = exc_info.value.args[0]
    assert exc_msg == "Channel number must be given as an integer."


def test_channel_tag(init_lst_proc):
    """Get / set the number of the channel the tag is in."""
    channel_number = 42
    init_lst_proc.channel_tag = channel_number
    assert init_lst_proc.channel_tag == channel_number


def test_channel_tag_wrong_type(init_lst_proc):
    """Raise a type error when a wrong type is set for the channel number."""
    wrong_type = "42"
    with pytest.raises(TypeError) as exc_info:
        init_lst_proc.channel_tag = wrong_type
    exc_msg = exc_info.value.args[0]
    assert exc_msg == "Channel number must be given as an integer."


def test_data_format_invalid(init_lst_proc):
    """Raise TypeError if an invalid data format is selected."""
    with pytest.raises(TypeError):
        init_lst_proc.data_format = "wrong type"


def test_file_name(init_lst_proc):
    """Get / set a filename and path."""
    test_path = Path("./test.txt")
    init_lst_proc.file_name = test_path
    assert init_lst_proc.file_name == test_path


def test_file_name_wrong_type(init_lst_proc):
    """Raise a TypeError when wrong file tpe was given."""
    wrong_type = 42
    with pytest.raises(TypeError) as exc_info:
        init_lst_proc.file_name = wrong_type
    exc_msg = exc_info.value.args[0]
    assert (
        exc_msg == f"Path must be a `pathlib.Path` object but is a {type(wrong_type)}."
    )


# METHODS #


def test_read_data_no_channel_data(init_lst_proc):
    """Raise ValueError if data channel is not set."""
    test_path = Path("./test.txt")
    init_lst_proc.file_name = test_path
    with pytest.raises(ValueError) as err:
        init_lst_proc.read_list_file()

    msg = err.value.args[0]
    assert msg == "Please set a number for the data channel."


def test_read_data_no_filename(init_lst_proc):
    """Raise ValueError if no file name provided."""
    with pytest.raises(ValueError) as err:
        init_lst_proc.read_list_file()

    msg = err.value.args[0]
    assert msg == "Please set a file name."


def test_set_data_format(init_lst_proc):
    """Set the data format according to dictionary values."""
    init_lst_proc._file_info["data_type"] = "asc"
    init_lst_proc._file_info["time_patch"] = "9"
    init_lst_proc.set_data_format()
    assert init_lst_proc._data_format == init_lst_proc.ASCIIFormat.ASC_9


def test_set_data_format_manually(init_lst_proc):
    """Set the data format according to dictionary values."""
    fmt = init_lst_proc.ASCIIFormat.ASC_9
    init_lst_proc.data_format = fmt
    assert init_lst_proc._data_format == fmt


def test_set_data_format_invalid_format(init_lst_proc):
    """Raise ValueError if an invalid format is provided."""
    data_type = "invalid"

    init_lst_proc._file_info["data_type"] = data_type
    init_lst_proc._file_info["time_patch"] = "9"

    with pytest.raises(ValueError) as err:
        init_lst_proc.set_data_format()

    msg = err.value.args[0]
    assert f"The data type {data_type.upper()}" in msg


def test_write_crd_no_data(init_lst_proc):
    """Raise ValueError if data has not been read yet."""
    with pytest.raises(ValueError) as err_info:
        init_lst_proc.write_crd()
    err_msg = err_info.value.args[0]
    assert err_msg == "No data has been read in yet."



"""Unit tests for the interfacer."""

from pathlib import Path

import pytest
import numpy as np

from rimseval import CRDFileProcessor, interfacer


def test_read_lion_eval_calfile(legacy_files_path, crd_file):
    """Read and apply an old LIONEval calibration file."""
    cal_fname = Path(legacy_files_path).joinpath("all_cals.cal")

    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))

    interfacer.read_lion_eval_calfile(crd, cal_fname)

    # make sure none of them are `None`
    assert crd.def_mcal is not None
    assert crd.def_integrals is not None
    assert crd.def_backgrounds is not None


def test_read_lion_eval_calfile_no_filename(crd_file):
    """Provide no filename will create one and error out because it does not exist."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))

    with pytest.raises(IOError) as err:
        interfacer.read_lion_eval_calfile(crd)

    err_msg = err.value.args[0]
    assert (
        err_msg == f"The requested calibration file {Path(fname).with_suffix('.cal')} "
        f"does not exist."
    )


def test_calfile_load_io_error(crd_file):
    """Raise IOError if file does not exist."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))

    fname = Path("does_not_exist")

    with pytest.raises(IOError) as err:
        interfacer.load_cal_file(crd, fname)

    err_msg = err.value.args[0]
    assert err_msg == f"The requested calibration file {fname} does not exist."


def test_calfile_save_reapply_cals(legacy_files_path, crd_file, tmpdir):
    """Save a calibration, apply to a second file, and assure they are identical."""
    _, _, _, fname = crd_file

    # first CRD file that we are going to open
    crd_1 = CRDFileProcessor(Path(fname))

    lion_cal = Path(legacy_files_path).joinpath("all_cals.cal")  # contains it all!
    interfacer.read_lion_eval_calfile(crd_1, lion_cal)

    settings_file = tmpdir.join("settings.json")  # temporary settings file
    interfacer.save_cal_file(crd_1, settings_file)

    # now open a second one and apply the same settings file
    crd_2 = CRDFileProcessor(Path(fname))
    interfacer.load_cal_file(crd_2, settings_file)

    # ensure they are equal
    np.testing.assert_equal(crd_1.def_mcal, crd_2.def_mcal)

    assert crd_1.def_integrals[0] == crd_2.def_integrals[0]
    np.testing.assert_equal(crd_1.def_integrals[1], crd_2.def_integrals[1])

    assert crd_1.def_backgrounds[0] == crd_2.def_backgrounds[0]
    np.testing.assert_equal(crd_1.def_backgrounds[1], crd_2.def_backgrounds[1])


def test_calfile_save_reapply_filters(legacy_files_path, crd_file, tmpdir):
    """Save a calibration, apply to a second file, and assure they are identical."""
    _, _, _, fname = crd_file

    # first CRD file that we are going to open
    crd_1 = CRDFileProcessor(Path(fname))

    # set some filters
    app_filters_exp = {
        "dead_time_corr": [True, 8],
        "max_ions_per_pkg": [True, 101],
        "max_ions_per_shot": [False, 1],
        "max_ions_per_time": [True, 11, 100.25],
        "max_ions_per_tof_window": [False, 11, [10.125, 20.125]],
        "pkg_peirce_rejection": True,
        "packages": [False, 1001],
        "spectrum_part": [True, [[2, 2], [6, 10]]],
    }
    crd_1.applied_filters = app_filters_exp

    settings_file = tmpdir.join("settings.json")  # temporary settings file
    interfacer.save_cal_file(crd_1, settings_file)

    # now open a second one and apply the same settings file
    crd_2 = CRDFileProcessor(Path(fname))
    interfacer.load_cal_file(crd_2, settings_file)

    # ensure they are equal
    assert crd_1.applied_filters == crd_2.applied_filters


def test_calfile_save_reapply_mcal_only(legacy_files_path, crd_file, tmpdir):
    """Save a mass calibration to default, reload again, and ensure it works."""
    _, _, _, fname = crd_file

    # first CRD file that we are going to open
    crd_1 = CRDFileProcessor(Path(fname))
    crd_1.def_mcal = np.array([[1, 1], [2, 2]])

    interfacer.save_cal_file(crd_1)
    interfacer.load_cal_file(crd_1)



"""Test error messages of the processor."""

from pathlib import Path

import numpy as np
import pytest

from rimseval import CRDFileProcessor


def test_dead_time_correction_no_shots(crd_file):
    """Raise warning if no shots were present in the crd file."""
    _, _, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))
    crd.nof_shots = 0

    msg_exp = "No data available"

    with pytest.warns(UserWarning, match=msg_exp):
        crd.dead_time_correction(10)


def test_filter_max_ions_per_pkg_smaller_one(crd_file):
    """Raise ValueError if max ions per pkg filter is smaller than 1."""
    _, _, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))

    err_exp = "The maximum number of ions must be larger than 1."

    with pytest.raises(ValueError) as err:
        crd.filter_max_ions_per_pkg(0)

    msg = err.value.args[0]
    assert err_exp == msg


def test_filter_max_ions_per_pkg_no_pkg_data(crd_file):
    """Raise OSError if no packaged data is available."""
    _, _, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))

    err_exp = "There is no packaged data."

    with pytest.raises(OSError) as err:
        crd.filter_max_ions_per_pkg(1)

    msg = err.value.args[0]
    assert err_exp in msg


def test_filter_max_ions_per_shot_smaller_one(crd_file):
    """Raise ValueError if number of shots is smaller than 1."""
    _, _, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))

    err_exp = "The maximum number of ions"

    with pytest.raises(ValueError) as err:
        crd.filter_max_ions_per_shot(0)

    msg = err.value.args[0]
    assert err_exp in msg


def test_filter_max_ions_per_tof_window(crd_file):
    """Raise ValueError if ToF window of wrong shape."""
    _, _, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))

    err_exp = "ToF window must be specified with two entries"

    with pytest.raises(ValueError) as err:
        crd.filter_max_ions_per_tof_window(10, np.array([13, 54, 199]))

    msg = err.value.args[0]
    assert err_exp in msg


def test_integral_windows_no_integrals_set(crd_file):
    """Raise ValueError if no integrals were set."""
    _, _, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))

    err_exp = "No integrals were set"

    with pytest.raises(ValueError) as err:
        crd.integrals_calc()

    msg = err.value.args[0]
    assert err_exp in msg


def test_integral_windows_no_mass(crd_file):
    """Raise ValueError if no mass calibration was applied."""
    _, _, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))
    peak_names = ["54Fe", "64Ni"]
    peak_limits = np.array([[53.8, 54.2], [63.5, 64.5]])
    crd.def_integrals = (peak_names, peak_limits)

    err_exp = "A mass calibration needs to be applied first"

    with pytest.raises(ValueError) as err:
        crd.integrals_calc()

    msg = err.value.args[0]
    assert err_exp in msg


def test_integral_calc_delta_no_integrals_set(crd_file):
    """Raise ValueError if no integrals were set."""
    _, _, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))

    err_exp = "No integrals were defined"

    with pytest.raises(ValueError) as err:
        crd.integrals_calc_delta()

    msg = err.value.args[0]
    assert err_exp in msg


def test_mass_calibration_no_params_set(crd_file):
    """Raise ValueError if mcal parameters were set."""
    _, _, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))

    err_exp = "No mass calibration was set"

    with pytest.raises(ValueError) as err:
        crd.mass_calibration()

    msg = err.value.args[0]
    assert err_exp in msg


def test_packages_wrong_shots(crd_file):
    """Raise ValueError if number of shots is too small or too large."""
    _, _, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))

    err_exp = "Number of shots per package must be between"

    with pytest.raises(ValueError) as err:
        crd.packages(0)

    msg = err.value.args[0]
    assert err_exp in msg


@pytest.mark.parametrize(
    "limit",
    [
        [[0, 3], "Your lower index"],
        [[1, 6], "Your upper index"],
        [[-1, 10], "lower and upper"],
    ],
)
def test_spectrum_part_range_index_error(limit, crd_file):
    """Raise ValueError range is in incorrect order."""
    _, _, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))

    with pytest.raises(IndexError) as err:
        crd.spectrum_part([limit[0]])

    msg = err.value.args[0]
    assert limit[1] in msg


def test_spectrum_part_range_wrong_order(crd_file):
    """Raise ValueError range is in incorrect order."""
    _, _, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))

    err_exp = "such that `from` < `to`"

    with pytest.raises(ValueError) as err:
        crd.spectrum_part([[4, 2]])

    msg = err.value.args[0]
    assert err_exp in msg


def test_spectrum_part_range_not_mutually_exclusive(crd_file):
    """Raise ValueError ranges are not mutually exclusive."""
    _, _, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))

    err_exp = "Your ranges are not mutually exclusive"

    with pytest.raises(ValueError) as err:
        crd.spectrum_part([[1, 3], [3, 4]])

    msg = err.value.args[0]
    assert err_exp in msg



"""Write Excel Files from the files that we have, e.g., a workup file."""

from datetime import datetime
from pathlib import Path

from iniabu.utilities import item_formatter
import xlsxwriter
from xlsxwriter.utility import xl_rowcol_to_cell

from .. import CRDFileProcessor
from ..utilities import ini


def workup_file_writer(
    crd: CRDFileProcessor, fname: Path, timestamp: bool = False
) -> None:
    """Write out an Excel workup file.

    This is for the user to write out an excel workup file, which will already be
    filled with the integrals of the given CRD file.

    :param crd: CRD file processor file to write out.
    :param fname: File name for the file to write out to.
    :param timestamp: Create a column for the time stamp? Defaults to ``False``
    """
    if crd.def_integrals is None:
        return
    else:
        int_names, _ = crd.def_integrals

    # format names
    for it, name in enumerate(int_names):
        formatted_name = item_formatter(name)
        if "-" in formatted_name:  # only format if it's an isotope!
            int_names[it] = item_formatter(name)

    # get normalizing isotope
    norm_names = []
    for name in int_names:
        try:
            ele = name.split("-")[0]
            norm_names.append(ini._get_norm_iso(ele))
        except IndexError:  # norm isotope does not exist
            norm_names.append(None)

    fname = fname.with_suffix(".xlsx").absolute()  # ensure correct format

    wb = xlsxwriter.Workbook(str(fname))
    ws = wb.add_worksheet()

    # formats
    fmt_title = wb.add_format({"bold": True, "font_size": 14})
    fmt_hdr_0 = wb.add_format({"bold": True, "italic": True, "left": True})
    fmt_hdr = wb.add_format({"bold": True, "italic": True})
    fmt_std_abus = wb.add_format({"bold": True, "color": "red"})
    fmt_counts_0 = wb.add_format({"num_format": "0", "left": True})
    fmt_counts = wb.add_format({"num_format": "0"})
    fmt_counts_unc = wb.add_format(
        {"num_format": "0", "color": "gray", "valign": "left"}
    )
    fmt_delta_0 = wb.add_format({"num_format": "0", "left": True})
    fmt_delta = wb.add_format({"num_format": "0"})
    fmt_delta_unc = wb.add_format(
        {"num_format": "0", "color": "gray", "valign": "left"}
    )
    fmt_timestamp = wb.add_format({"num_format": "YYYY-MM-DD HH:MM:SS"})

    # cell widths for data
    wdth_counts = 8
    wdth_delta = 12
    wdth_delta_unc = 6

    # rows to define
    abu_row = 4
    num_eqn_rows = 999

    # special headers and user definitions
    general_headers = ["Remarks", "File Name", "# Shots"]
    general_headers_widths = [20, 20, 10]
    fname_col = 1
    shots_col = 2

    if timestamp:
        general_headers.append("Timestamp")
        general_headers_widths.append(18)

    # write the title
    ws.write(0, 0, f"Workup {datetime.today().date()}", fmt_title)

    # write data header and abundances
    hdr_row = abu_row + 1  # row to start header of the data in

    int_col = len(general_headers)  # start of integral column
    delta_col = int_col + 2 * len(int_names)  # start of delta column

    for col, hdr in enumerate(general_headers):
        ws.write(hdr_row, col, hdr, fmt_hdr)
        ws.set_column(col, col, general_headers_widths[col])

    for col, name in enumerate(int_names):
        # write abundances
        try:
            abu_col = ini.iso[name].abu_rel
            ws.write(abu_row, 2 * col + int_col, abu_col, fmt_std_abus)
        except IndexError:
            pass
        # write integral header
        name = iso_format_excel(name)

        fmt_hdr_use = fmt_hdr_0 if col == 0 else fmt_hdr

        ws.write(hdr_row, 2 * col + int_col, name, fmt_hdr_use)
        ws.write(hdr_row, 2 * col + 1 + int_col, "±1σ", fmt_hdr)

    # integral column width
    ws.set_column(int_col, int_col + 2 * len(int_names) - 1, wdth_counts)

    col = delta_col
    for it, name in enumerate(int_names):
        norm_iso_name = norm_names[it]
        fmt_hdr_use = fmt_hdr_0 if col == delta_col else fmt_hdr
        if (
            norm_iso_name is not None and norm_iso_name in int_names
        ):  # norm isotope valid
            ws.write(
                hdr_row,
                col,
                f"δ({iso_format_excel(name)}/{iso_format_excel(norm_iso_name)})",
                fmt_hdr_use,
            )
            ws.write(hdr_row, col + 1, "±1σ", fmt_hdr)

            # set width
            ws.set_column(col, col, wdth_delta)
            ws.set_column(col + 1, col + 1, wdth_delta_unc)

            col += 2

    # WRITE DATA
    data_row = hdr_row + 1

    # file naame
    ws.write(data_row, fname_col, crd.fname.name)

    # write the number of shots
    ws.write(data_row, shots_col, crd.nof_shots, fmt_counts)

    for col in range(data_row + 1, num_eqn_rows + data_row):
        ws.write_blank(col, shots_col, None, fmt_counts)

    # write the timestamp if requested
    if timestamp:
        ws.write(data_row, shots_col + 1, crd.timestamp, fmt_timestamp)

    for col in range(data_row + 1, num_eqn_rows + data_row):
        ws.write_blank(col, shots_col + 1, None, fmt_timestamp)

    # write integrals
    for col, dat in enumerate(crd.integrals):
        fmt_counts_use = fmt_counts_0 if col == 0 else fmt_counts
        ws.write(data_row, 2 * col + int_col, dat[0], fmt_counts_use)
        ws.write(data_row, 2 * col + int_col + 1, dat[1], fmt_counts_unc)

    for col in range(data_row + 1, num_eqn_rows + data_row):  # boarder for integrals
        ws.write_blank(col, int_col, None, fmt_counts_0)

    # write delta equations
    col = delta_col
    for it, _ in enumerate(int_names):
        norm_iso_name = norm_names[it]
        if (
            norm_iso_name is not None and norm_iso_name in int_names
        ):  # norm isotope valid
            for eq_row in range(num_eqn_rows):
                # get cell values, nominators and denominators
                nom_iso = xl_rowcol_to_cell(data_row + eq_row, 2 * it + int_col)
                nom_iso_unc = xl_rowcol_to_cell(data_row + eq_row, 2 * it + int_col + 1)
                den_iso = xl_rowcol_to_cell(
                    data_row + eq_row,
                    2 * int_names.index(norm_iso_name) + int_col,
                    col_abs=True,
                )
                den_iso_unc = xl_rowcol_to_cell(
                    data_row + eq_row,
                    2 * int_names.index(norm_iso_name) + int_col + 1,
                    col_abs=True,
                )
                # get standard abundances cell names
                nom_std = xl_rowcol_to_cell(abu_row, 2 * it + int_col, row_abs=True)
                den_std = xl_rowcol_to_cell(
                    abu_row,
                    2 * int_names.index(norm_iso_name) + int_col,
                    row_abs=True,
                    col_abs=True,
                )

                # decide if we write a boarder or not
                fmt_delta_use = fmt_delta_0 if col == delta_col else fmt_delta

                # write the values for the delta formula
                ws.write(
                    data_row + eq_row,
                    col,
                    f'=IF({nom_iso}<>"",'
                    f'(({nom_iso}/{den_iso})/({nom_std}/{den_std})-1)*1000, "")',
                    fmt_delta_use,
                )
                # equation for uncertainty
                ws.write(
                    data_row + eq_row,
                    col + 1,
                    f'=IF({nom_iso}<>"",'
                    f"1000*SQRT("
                    f"(({nom_iso_unc}/{den_iso})/({nom_std}/{den_std}))^2+"
                    f"(({nom_iso}*{den_iso_unc}/{den_iso}^2)/({nom_std}/{den_std}))^2"
                    f'), "")',
                    fmt_delta_unc,
                )

            # increment column
            col += 2

    # close the workbook
    wb.close()


def iso_format_excel(iso: str) -> str:
    """Format isotope name from `iniabu` to format as written to Excel.

    :param iso: Isotope, formatted according to `iniabu`, e.g., "Si-28"
    :return: Excel write-out format.
    """
    iso_split = iso.split("-")
    if len(iso_split) != 2:
        return iso
    else:
        return f"{iso_split[1]}{iso_split[0]}"



"""Run Pierce's criterion to reject data.

Implementation after Ross (2003) using calculation method for table from Wikipedia.
Note that the table that Ross (2003) presents is for `R`, which is the square root
of what `x**2` means in Gould (1855). Also, the first value of Ross (2003) for
three observations, one doubtful value seems to off by a bit. The rest of the
table agrees well.
"""

from typing import Tuple

import numpy as np
from scipy import special


def peirce_criterion(n_tot: int, n: int, m: int = 1) -> float:
    """Peirce's criterion.

    Returns the threshold error deviation for outlier identification
    using Peirce's criterion based on Gould's methodology.
    This routine is heavily copied from Wikipedia

    :param n_tot: Total number of observations.
    :param n: Number of outliers to be removed.
    :param m: Number of model unknowns, defaults to 1.

    :return: Error threshold `R` (Ross, 2003) / Square root of `x**2` (Gould, 1955)
    """
    # Check number of observations:
    if n_tot > 1:
        # Calculate Q (Nth root of Gould's equation B):
        q_cap = (n ** (n / n_tot) * (n_tot - n) ** ((n_tot - n) / n_tot)) / n_tot

        # Initialize R values (as floats)
        r_new = 1.0
        r_old = 0.0
        #
        # Start iteration to converge on R:
        while abs(r_new - r_old) > (n_tot * 2.0e-16):
            # Calculate Lamda
            # (1/(N-n)th root of Gould's equation A'):
            ldiv = r_new**n
            if ldiv == 0:
                ldiv = 1.0e-6
            lambda_g = ((q_cap**n_tot) / (ldiv)) ** (1.0 / (n_tot - n))
            # Calculate x-squared (Gould's equation C):
            x2 = 1.0 + (n_tot - m - n) / n * (1.0 - lambda_g**2.0)
            # If x2 goes negative, return 0:
            if x2 < 0:
                x2 = 0.0
                r_old = r_new
            else:
                # Use x-squared to update R (Gould's equation D):
                r_old = r_new
                r_new = np.exp((x2 - 1) / 2.0) * special.erfc(
                    np.sqrt(x2) / np.sqrt(2.0)
                )
    else:
        x2 = 0.0
    return np.sqrt(x2)


def reject_outliers(
    data: np.ndarray, m: int = 1
) -> Tuple[float, float, np.ndarray, np.ndarray]:
    """Apply Peirce's criterion to reject outliers.

    Algorithm implmeneted as given by Ross (2003).

    :param data: All data points.
    :param m: Number of model unknowns, defaults to 1.

    :return: New average and standard deviation, Array with the outliers.
    """
    data = np.array(data)  # just making sure it's a numpy array
    data_in = data.copy()

    # first drop all nan and inf
    drop_index = np.where(np.isnan(data) | np.isinf(data))
    data = np.delete(data, drop_index)

    avg = np.average(data)
    std = np.std(data)
    n_tot = len(data)

    outliers = []
    diffs = np.abs(data - avg)

    for it in range(len(data)):  # check for every data point if it should be rejected
        max_diff = diffs.max()
        max_ind = diffs.argmax()

        rejection_limit = peirce_criterion(n_tot, it + 1, m)
        if max_diff > rejection_limit * std:
            outliers.append(data[max_ind])
            # delete max from diffs and data
            data = np.delete(data, max_ind)
            diffs = np.delete(diffs, max_ind)
        else:
            break  # we are done rejecting

    avg_new = float(np.average(data))
    std_new = float(np.std(data))

    # determine indexes of outliers in original array
    outliers = np.array(outliers)
    outliers_ind = np.zeros_like(outliers)
    for it, outl in enumerate(outliers):
        outliers_ind[it] = np.where(data_in == outl)[0][0]  # only first occurence
    outliers_ind = np.sort(outliers_ind)

    return avg_new, std_new, outliers, outliers_ind



"""This file contains utilities that do not fit anywhere else."""

import iniabu
from numba import njit
import numpy as np

ini = iniabu.IniAbu(database="nist")  # select correct iniabu database for this program


@njit
def not_index(ind: np.array, length: int) -> np.array:  # pragma: nocover
    """Reverse an index.

    After filtering ions, e.g., using `np.where`, and keeping the indices of an array,
    this routine creates the opposite, i.e., an array of the indices that have not been
    filtered.

    :param ind: Array of all the indexes.
    :param length: Length of the original index to take out of.

    :return: The reversed index.

    :raises ValueError: Max index is larger than the total length, which should not be.

    Example:
        >>> a = np.arange(7)
        >>> b = np.where(a < 4)[0]
        >>> b
        array([0, 1, 2, 3])
        >>> not_index(b)
        array([4, 5, 6])
    """
    if ind.shape == (0,):  # no indices filtered out
        return np.arange(length)
    else:
        if np.max(ind) >= length:
            raise ValueError("The maximum index must be smaller than the length.")

        not_arr = np.array([it for it in range(length) if it not in ind])
        return not_arr



"""Test the macro capability."""

from pathlib import Path

import pytest
import numpy as np

from rimseval.processor import CRDFileProcessor


@pytest.fixture
def macro_path(request) -> Path:
    """Provides the path to this folder where the macros are.

    :return: Path to the folder
    """
    return Path(request.fspath).parents[0].absolute()


def test_macro_max_ions_per_shot(macro_path, crd_file):
    """Run macro to filter maximum ions per shot and assert it's functioning."""
    _, _, _, fname = crd_file

    crd_exp = CRDFileProcessor(Path(fname))
    crd_exp.spectrum_full()
    crd_exp.filter_max_ions_per_shot(3)

    macro_file = macro_path.joinpath("ex_max_ions_per_shot.py")

    crd_macro = CRDFileProcessor(Path(fname))
    crd_macro.spectrum_full()
    crd_macro.run_macro(macro_file)

    np.testing.assert_equal(crd_exp.data, crd_macro.data)
    assert crd_exp.nof_shots == crd_macro.nof_shots


def test_macro_max_ions_per_pkg(macro_path, crd_file):
    """Run macro to filter out packages with too many shots and compare to filter."""
    _, _, _, fname = crd_file

    crd_exp = CRDFileProcessor(Path(fname))
    crd_exp.spectrum_full()
    crd_exp.packages(2)
    crd_exp.filter_max_ions_per_pkg(4)

    macro_file = macro_path.joinpath("ex_max_ions_per_pkg.py")

    crd_macro = CRDFileProcessor(Path(fname))
    crd_macro.spectrum_full()
    crd_macro.packages(2)
    crd_macro.run_macro(macro_file)

    np.testing.assert_equal(crd_exp.data, crd_macro.data)
    assert crd_exp.nof_shots == crd_macro.nof_shots
    np.testing.assert_equal(crd_exp.data_pkg, crd_macro.data_pkg)
    np.testing.assert_equal(crd_exp.nof_shots_pkg, crd_macro.nof_shots_pkg)



"""Function test for processor class methods, focusing on each function."""

import datetime
from pathlib import Path

import pytest
import numpy as np

from rimseval.processor import CRDFileProcessor
import rimseval.processor_utils as pu

# TEST PROPERTIES #


def test_timestamp(crd_file):
    """Get the time stamp of the CRD file as a python datetime."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))
    timestamp = crd.timestamp
    assert isinstance(timestamp, datetime.datetime)


# TEST METHODS #


def test_data_dimension_after_dead_time_correction(crd_file):
    """Ensure ToF and data have the same dimensions - BF 2021-07-23."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.dead_time_correction(3)

    assert crd.tof.ndim == crd.data.ndim


def test_filter_max_ions_per_pkg(crd_file):
    """Filter the packages by maximum ion."""
    _, ions_per_shot, _, fname = crd_file
    max_ions = ions_per_shot.max() - 1  # filter one or so packages out
    sum_ions = 0
    for ion in ions_per_shot:
        if ion <= max_ions:
            sum_ions += ion

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.packages(1)
    crd.filter_max_ions_per_pkg(max_ions)
    assert crd.data_pkg.sum() == sum_ions


def test_filter_max_ions_per_shot(crd_file):
    """Filter the shots by maximum ions per shot."""
    _, ions_per_shot, _, fname = crd_file
    max_ions = ions_per_shot.min() + 1  # filter most out
    filtered_data = ions_per_shot[np.where(ions_per_shot <= max_ions)]
    sum_ions_exp = np.sum(filtered_data)
    nof_shots_exp = len(filtered_data)

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.filter_max_ions_per_shot(max_ions)

    assert crd.nof_shots == nof_shots_exp
    assert crd.data.sum() == sum_ions_exp
    np.testing.assert_equal(crd.ions_per_shot, filtered_data)


def test_filter_max_ions_per_shot_double(crd_file):
    """Test filterting max ions per shot twice (no pkgs)."""
    header, ions_per_shot, all_tofs, fname = crd_file
    max_ions1 = max(ions_per_shot) - 1  # filter the highest one out
    max_ions2 = min(ions_per_shot) + 1

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.filter_max_ions_per_shot(max_ions1)
    crd.filter_max_ions_per_shot(max_ions2)

    ions_per_shot_filtered = ions_per_shot[np.where(ions_per_shot <= max_ions2)]
    nof_shots = len(ions_per_shot_filtered)
    nof_ions = np.sum(ions_per_shot_filtered)

    assert crd.nof_shots == nof_shots
    assert np.sum(crd.data) == nof_ions


def test_filter_max_ions_per_shot_pkg(crd_file):
    """Test maximum ions per shot filtering with packages."""
    header, ions_per_shot, all_tofs, fname = crd_file
    max_ions = max(ions_per_shot) - 1  # filter the highest one out
    shots_per_pkg = 2

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.packages(shots_per_pkg)
    crd.filter_max_ions_per_shot(max_ions)

    # assert that package data are the same as the rest
    assert crd.nof_shots == crd.nof_shots_pkg.sum()
    assert crd.data.sum() == crd.data_pkg.sum()


def test_filter_max_ions_per_shot_pkg_filtered(crd_file):
    """Test maximum ions per shot filtering with packages and pkg filter applied."""
    header, ions_per_shot, all_tofs, fname = crd_file
    shots_per_pkg = 2
    max_ions = 1
    max_ions_per_pkg = 4

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.packages(shots_per_pkg)
    crd.filter_max_ions_per_pkg(max_ions_per_pkg)
    crd.filter_max_ions_per_shot(max_ions)

    # assert that package data are the same as the rest
    assert crd.nof_shots == crd.nof_shots_pkg.sum()
    assert crd.data.sum() == crd.data_pkg.sum()


def test_filter_max_ions_per_time(crd_file):
    """Test maximum ions per shot in given time window."""
    header, ions_per_shot, all_tofs, fname = crd_file
    max_ions = 1  # filter the highest one out
    time_window_us = 39 * 100 / 1e6  # 40 channels, filters third but not fifth

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.filter_max_ions_per_time(max_ions, time_window_us)

    assert crd.nof_shots == len(ions_per_shot) - 1
    assert np.sum(crd.data) == np.sum(ions_per_shot) - 4


def test_filter_max_ions_per_time_nothing(crd_file):
    """Do nothing if filter shows no entries for removal."""
    header, ions_per_shot, all_tofs, fname = crd_file
    max_ions = ions_per_shot.max()  # no ions will be filtered
    time_window_us = 80e-6  # channel width for MCS8a

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.filter_max_ions_per_time(max_ions, time_window_us)

    assert crd.nof_shots == len(ions_per_shot)
    assert np.sum(crd.data) == np.sum(ions_per_shot)


def test_filter_max_ions_per_tof_window(crd_file):
    """Test maximum ions per shot in given time window."""
    header, ions_per_shot, all_tofs, fname = crd_file
    max_ions = 1  # filter the highest one out
    tof_window_us = (
        np.array([221, 281]) * 100 / 1e6
    )  # filter out the last one, but none of the others

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.filter_max_ions_per_tof_window(max_ions, tof_window_us)

    assert crd.nof_shots == len(ions_per_shot) - 1
    assert np.sum(crd.data) == np.sum(ions_per_shot) - 2


def test_filter_max_ions_per_tof_window_conv_tof_window_no_shots(crd_file):
    """Convert ToF window to np.ndarry, no filtering since no ions in criterion."""
    header, ions_per_shot, all_tofs, fname = crd_file
    max_ions = ions_per_shot.sum()  # filter the highest one out
    tof_window_us = (221, 400)
    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.filter_max_ions_per_tof_window(max_ions, tof_window_us)

    assert crd.nof_shots == len(ions_per_shot)
    assert np.sum(crd.data) == np.sum(ions_per_shot)


def test_mass_calibration_2pts(crd_file):
    """Perform mass calibration with two points."""
    _, _, _, fname = crd_file
    params = (13, 42)
    tms = (42.0, 95.0)

    mass_cal = np.zeros((len(tms), 2))
    for it, tm in enumerate(tms):
        mass_cal[it][0] = tm
        mass_cal[it][1] = pu.tof_to_mass(tm, params[0], params[1])

    # set variables
    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()

    crd.def_mcal = mass_cal
    mass_exp = pu.tof_to_mass(crd.tof, params[0], params[1])

    crd.mass_calibration()
    mass_rec = crd.mass
    print(tms)
    np.testing.assert_almost_equal(mass_rec, mass_exp)
    assert crd.mass.ndim == crd.tof.ndim


@pytest.mark.parametrize("new_integral", [None, (["Int2"], np.array([[3.0, 4.0]]))])
def test_integrals_definition_delete_undefined_background_to_none(
    crd_file, new_integral
):
    """Delete backgrounds and set to none if peak goes to undefined."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))

    crd.def_integrals = ["Integral"], np.array([[1.0, 2.0]])
    crd.def_backgrounds = ["Integral"], np.array([[2.0, 3.0]])
    crd.def_integrals = new_integral
    assert crd.def_backgrounds is None


def test_integrals_definition_delete_undefined_background(crd_file):
    """Delete backgrounds and set to none if peak goes to undefined."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))

    crd.def_integrals = ["Int1", "Int2"], np.array([[1.0, 2.0], [2.0, 3.0]])
    crd.def_backgrounds = ["Int1", "Int2"], np.array([[2.0, 3.0], [4.0, 5.0]])
    crd.def_integrals = ["Int1"], np.array([[1.0, 2.0]])

    assert "Int2" not in crd.def_backgrounds[0]


def test_integrals_delta_calc(crd_int_delta):
    """Assure that the delta calculation on the integral returns correct values."""
    crd_int_delta.integrals_calc_delta()
    deltas_exp = np.array([0, 0, 0, 0, np.nan, np.nan, np.nan])
    deltas_rec = crd_int_delta.integrals_delta.transpose()[0]  # test only deltas
    np.testing.assert_almost_equal(deltas_exp, deltas_rec)
    for line in crd_int_delta.integrals_pkg:
        np.testing.assert_almost_equal(crd_int_delta.integrals, line)


def test_packages(crd_file):
    """Simple test to ensure packages are made in two ways correctly."""
    _, ions_per_shot, _, fname = crd_file
    nof_shots = len(ions_per_shot)
    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.packages(nof_shots // 2)

    assert crd.data_pkg.sum() == crd.data.sum()
    np.testing.assert_equal(crd.data_pkg.sum(axis=0), crd.data)
    assert crd.nof_shots_pkg.sum() == crd.nof_shots
    # now redo w/ a lower number of shots per pkg
    crd.packages(nof_shots // 4)
    assert crd.data_pkg.sum() == crd.data.sum()
    np.testing.assert_equal(crd.data_pkg.sum(axis=0), crd.data)
    assert crd.nof_shots_pkg.sum() == crd.nof_shots


def test_sort_integrals(crd_file):
    """Sort integrals by first mass."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))
    crd.sort_integrals()  # does nothing, since None
    crd.def_integrals = ["Fe-56", "Ti-46"], np.array([[55.8, 56.2], [45.8, 46.2]])

    crd.integrals = np.array([[2, 2], [1, 1]])
    crd.integrals_pkg = np.array([crd.integrals, crd.integrals - 1])

    names_exp = ["Ti-46", "Fe-56"]
    values_exp = np.array([[45.8, 46.2], [55.8, 56.2]])
    integrals_exp = np.array([[1, 1], [2, 2]])
    integrals_pkg_exp = np.array([[[1, 1], [2, 2]], [[0, 0], [1, 1]]])

    crd.sort_integrals()
    crd.sort_integrals()  # second time, this should do nothing!
    names_rec, values_rec = crd.def_integrals
    assert names_rec == names_exp
    np.testing.assert_equal(values_rec, values_exp)
    np.testing.assert_equal(crd.integrals, integrals_exp)
    np.testing.assert_equal(crd.integrals_pkg, integrals_pkg_exp)


def test_sort_integrals_primary_and_secondary(crd_file) -> None:
    """Sort integrals first by element (if available), then by mass.

    This is especially important when we have two elements that are shifted by half
    masses.

    :param crd_file: CRD test file.
    """
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))
    crd.sort_integrals()  # does nothing, since None
    crd.def_integrals = ["Zr-94", "Mo-96", "Mo-94", "Zr-96"], np.array(
        [[93.9, 94.2], [95.4, 95.7], [93.4, 93.7], [95.9, 96.2]]
    )
    crd.integrals = np.array([[0, 0], [3, 3], [2, 2], [1, 1]])

    names_exp = ["Zr-94", "Zr-96", "Mo-94", "Mo-96"]
    masses_exp = np.array([[93.9, 94.2], [95.9, 96.2], [93.4, 93.7], [95.4, 95.7]])
    integrals_exp = np.array([[0, 0], [1, 1], [2, 2], [3, 3]])

    crd.sort_integrals()
    names_rec, masses_rec = crd.def_integrals
    assert names_rec == names_exp
    assert masses_rec == pytest.approx(masses_exp)
    np.testing.assert_equal(crd.integrals, integrals_exp)


def test_sort_integrals_non_element(crd_file) -> None:
    """Sort recognizable elements and other names together, all others last."""
    _, _, _, fname = crd_file
    crd = CRDFileProcessor(Path(fname))
    crd.sort_integrals()  # does nothing, since None
    crd.def_integrals = ["ZrO-94", "MoO-96", "Mo-94", "Zr-96"], np.array(
        [[93.9, 94.2], [95.4, 95.7], [93.4, 93.7], [95.9, 96.2]]
    )
    crd.integrals = np.array([[0, 0], [3, 3], [2, 2], [1, 1]])

    names_exp = ["Zr-96", "Mo-94", "ZrO-94", "MoO-96"]
    crd.sort_integrals()
    assert crd.def_integrals[0] == names_exp


def test_spectrum_part(crd_file):
    """Cut spectrum by two shots."""
    _, ions_per_shot, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.spectrum_part([1, len(ions_per_shot) - 2])

    assert crd.nof_shots == len(ions_per_shot) - 2


def test_spectrum_part_no_tof_indexes(crd_file):
    """Cut spectrum by two shots."""
    _, ions_per_shot, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.spectrum_part([2, 2])

    assert crd.nof_shots == 1
    assert crd.data.sum() == 0


def test_spectrum_part_data_length(crd_file):
    """Ensure that the data length is not cut."""
    _, _, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.spectrum_part([1, 2])

    assert len(crd.data) == len(crd.tof)


def test_spectrum_part_undo(crd_file):
    """Cut spectrum by two shots."""
    _, ions_per_shot, _, fname = crd_file

    crd = CRDFileProcessor(Path(fname))
    crd.spectrum_full()
    crd.spectrum_part([1, len(ions_per_shot) - 2])
    # undo the spectrum_part
    crd.spectrum_full()

    assert crd.nof_shots == len(ions_per_shot)



"""Multi processor unit tests."""

from pathlib import Path

import pytest
import numpy as np

from rimseval.multi_proc import MultiFileProcessor as mfp


@pytest.mark.parametrize("params", [[True, False], [False, True]])
def test_mfp_apply_to_all(params, crd_file):
    """Apply settings from one crd file to all.

    :param params: Parameters for: bg_corr, opt_mcal
    :param crd_file: CRD file to process.
    """
    _, _, _, fname = crd_file
    files = [Path(fname), Path(fname)]
    id = 0  # id of main crd file
    bg_corr = params[0]

    crds = mfp(files)
    crds.read_files()

    crd_main = crds.files[id]
    crd_main.def_mcal = np.array([[0.1, 10], [0.2, 22], [0.3, 33]])
    crd_main.mass_calibration()
    crd_main.def_integrals = ["Int1"], np.array([[6.0, 7.0]])
    crd_main.def_backgrounds = ["Int1"], np.array([[5.0, 6.0]])
    crd_main.packages(2)
    crd_main.filter_max_ions_per_pkg(5)
    crd_main.integrals_calc(bg_corr=bg_corr)

    # apply to all
    crds.apply_to_all(id=id, opt_mcal=params[1], bg_corr=bg_corr)

    for crd in crds.files:
        np.testing.assert_equal(crd.data, crd_main.data)
        np.testing.assert_equal(crd.mass, crd_main.mass)
        np.testing.assert_equal(crd.tof, crd_main.tof)
        np.testing.assert_equal(crd.data_pkg, crd_main.data_pkg)
        assert crd.nof_shots == crd_main.nof_shots
        np.testing.assert_equal(crd.nof_shots_pkg, crd_main.nof_shots_pkg)
        assert crd.fname.with_suffix(".json").is_file()


def test_mfp_apply_to_all_main_file_not_processed(crd_file):
    """Apply settings from one crd file to all, while one file is not yet processed."""
    _, _, _, fname = crd_file
    files = [Path(fname), Path(fname)]
    id = 0  # id of main crd file
    bg_corr = True

    crds = mfp(files)
    crds.open_files()

    crd_main = crds.files[id]
    crd_main.def_mcal = np.array([[0.1, 10], [0.2, 22], [0.3, 33]])
    crd_main.def_integrals = ["Int1"], np.array([[6.0, 7.0]])
    crd_main.def_backgrounds = ["Int1"], np.array([[5.0, 6.0]])

    # apply to all
    crds.apply_to_all(id=id, opt_mcal=False, bg_corr=bg_corr)

    # assert main file was processed right
    assert crd_main.tof is not None
    assert crd_main.data is not None
    assert crd_main.mass is not None
    assert crd_main.integrals is not None

    for crd in crds.files:
        np.testing.assert_equal(crd.data, crd_main.data)
        np.testing.assert_equal(crd.mass, crd_main.mass)
        np.testing.assert_equal(crd.tof, crd_main.tof)
        np.testing.assert_equal(crd.data_pkg, crd_main.data_pkg)
        assert crd.nof_shots == crd_main.nof_shots
        np.testing.assert_equal(crd.nof_shots_pkg, crd_main.nof_shots_pkg)
        assert crd.fname.with_suffix(".json").is_file()


def test_mfp_load_calibrations(crd_file, crd_data):
    """Load calibrations with primary and secondary file."""
    _, _, _, fname = crd_file
    files = [Path(fname), crd_data.joinpath("ti_standard_01.crd")]

    secondary_cal_file = crd_data.joinpath("ti_standard_01.json")

    crds = mfp(files)
    crds.load_calibrations(secondary_cal=secondary_cal_file)

    for crd in crds.files:
        assert crd.def_mcal is not None  # loading complete



"""Create a schematic figure for Background Correction explanation."""

import matplotlib.pyplot as plt
import numpy as np

channels = np.arange(0, 1, 0.001)

# background
bg = np.random.rand(len(channels))

# a peak at 0.5
pk_rnd = np.random.randn(100000)
pk, _, _ = plt.hist(pk_rnd, bins=len(channels))
pk /= np.max(pk) / 10

# signal
signal = pk + bg

# areas to mark
bg_shades = [[0.05, 0.15], [0.85, 0.95]]
pk_shade = 0.25, 0.75

# plot
fig, ax = plt.subplots(1, 1)
fig.set_size_inches(4, 3)

ax.fill_between(
    channels, signal, np.zeros_like(channels), color="tab:blue", linewidth=0
)

# mark bg
bgcol = "tab:orange"
for it, lims in enumerate(bg_shades):  # areas must be in limits
    lim1 = np.where(channels >= lims[0])[0][0]
    lim2 = np.where(channels >= lims[1])[0][0]
    ax.fill_between(
        channels[lim1:lim2],
        signal[lim1:lim2],
        np.zeros_like(channels[lim1:lim2]),
        color=bgcol,
        linewidth=0,
    )
    ax.text(
        channels[lim2 - int((lim2 - lim1) / 2)],
        1.3,
        f"$B_{{{it+1}}}$",
        ha="center",
        va="bottom",
    )

# mark peak
lim1 = np.where(channels >= pk_shade[0])[0][0]
lim2 = np.where(channels >= pk_shade[1])[0][0]
ax.fill_between(
    channels[lim1:lim2],
    signal[lim1:lim2],
    np.zeros_like(channels[lim1:lim2]),
    color="tab:green",
    linewidth=0,
)
ax.text(0.6, 9, "$A_{p}$")

# remove ticks and markers
ax.xaxis.set_tick_params(labelbottom=False)
ax.yaxis.set_tick_params(labelleft=False)
ax.set_xticks([])
ax.set_yticks([])

# labels
ax.set_xlabel("Channel")
ax.set_ylabel("Signal")

fig.tight_layout()

# fig.show()
fig.savefig("bg_corr.png", dpi=300)



"""Contains legacy routines that allow interaction with old LIONEval cal files."""

from pathlib import Path
from typing import List, Union
import warnings

import numpy as np


class LIONEvalCal:
    """Reads a calibration file from the LIONEval program."""

    def __init__(self, fname: Path) -> None:
        """Read all the data in and get ready to return them with properties only.

        :param fname: File name for old calibration file

        :raises TypeError: fname is not an instance of Path
        """
        if not isinstance(fname, Path):
            raise TypeError("Filename must be given as a valid Path using pathlib.")
        self._fname = fname

        self._mass_cal = None
        self._integrals = None
        self._bg_corr = None
        self._applied_filters = None

        self._process_file()

    @property
    def applied_filters(self) -> dict:
        """Return the applied filters.

        :return: Applied filters as loaded from LION cal file, in new format.
        """
        return self._applied_filters

    @property
    def mass_cal(self) -> np.ndarray:
        """Return the mass calibration from the file, if present.

        :return: Mass calibration as ndarray. Columns as following:
            1st: ToF (us)
            2nd: Mass (amu)
        """
        return self._mass_cal

    @property
    def integrals(self) -> List[List[Union[str, float]]]:
        """Return the integrals from the file, if present.

        :return: Integrals. Columns as following:
            1st: Name of the peak.
            2nd: Mass of the center (amu).
            3rd: Interval from center of mass to lower bound (amu).
            4th: Interval from center of mass to upper bound (amu).
        """
        return self._integrals

    @property
    def bg_corr(self) -> List[List[Union[str, float]]]:
        """Return background correction data of the peak, if present.

        :return: Background correction. Columns as following:
            1st: Name of the peak.
            2nd: Mass of the center (amu).
            3rd: Lower bound of background left of peak (amu).
            4th: Upper bound of background left of peak (amu).
            5th: Lower bound of background right of peak (amu).
            6th: Upper bound of background right of peak (amu).
        """
        return self._bg_corr

    def _process_file(self) -> None:
        """Read in the calibration file and process the various properties."""
        with self._fname.open("r") as f:
            data_in = []
            for line in f:
                data_in.append(line.rstrip())

        mcal_block = _extract_block(data_in, top_char="# Mass Calibration")
        integral_block = _extract_block(data_in, top_char="# Integral Calibration")
        bg_block = _extract_block(data_in, top_char="# Background Calibration")
        settings_block = _extract_block(
            data_in, top_char="# recalculation settings", bott_char="# EOF"
        )

        self._read_and_set_mcal(mcal_block)
        self._read_and_set_integrals(integral_block)
        self._read_and_set_bg_corr(bg_block)
        self._read_and_set_settings_calculation(settings_block)

    def _read_and_set_mcal(self, mcal_block: List[str]) -> None:
        """Read, parse, and set a mass calibration block."""
        if not mcal_block:
            return None

        mcal = []
        for line in mcal_block:
            entry = line.split()
            mcal.append([float(entry[0]), float(entry[1])])

        self._mass_cal = np.array(mcal)

    def _read_and_set_integrals(self, int_block: List[str]) -> None:
        """Read, parse, and set a integral block."""
        if not int_block:
            return None

        integrals = []
        for line in int_block:
            entry = line.split()
            integrals.append(
                [entry[0], float(entry[1]), float(entry[2]), float(entry[3])]
            )

        self._integrals = integrals

    def _read_and_set_bg_corr(self, bg_block: List[str]) -> None:
        """Read, parse, and set a background correction block."""
        if not bg_block:
            return None

        bg_corr = []
        for line in bg_block:
            entry = line.split()
            bg_corr.append(
                [
                    entry[0],
                    float(entry[1]),
                    float(entry[2]),
                    float(entry[3]),
                    float(entry[4]),
                    float(entry[5]),
                ]
            )

        self._bg_corr = bg_corr

    def _read_and_set_settings_calculation(self, settings_block: List[str]) -> None:
        """Read adn parse the parameters block."""
        if not settings_block:
            return None

        if len(settings_block) != 21:
            warnings.warn(
                "Calibration file contains the wrong number of settings. "
                "No calculation filters will be loaded."
            )
            return None

        applied_filters = {
            "dead_time_corr": [
                bool(int(settings_block[0])),
                int(settings_block[1]),
            ],
            "packages": [
                bool(int(settings_block[2])),
                int(settings_block[3]),
            ],
            "max_ions_per_shot": [bool(int(settings_block[4])), int(settings_block[5])],
            "max_ions_per_pkg": [bool(int(settings_block[6])), int(settings_block[7])],
            "max_ions_per_time": [
                bool(int(settings_block[8])),
                int(settings_block[9]),
                float(settings_block[10]),
            ],
            "max_ions_per_tof_window": [
                bool(int(settings_block[11])),
                int(settings_block[12]),
                [float(settings_block[13]), float(settings_block[14])],
            ],
            "spectrum_part": [
                bool(int(settings_block[15])),
                [int(settings_block[16]), int(settings_block[17])],
            ],
        }

        self._applied_filters = applied_filters


def _extract_block(data: List[str], top_char: str, bott_char: str = "#") -> List[str]:
    """Extract a block from a given file.

    :param data: Data, i.e., the file content in a list with each line as an entry.
    :param top_char: Start of line above the block of interest (excluded)
    :param bott_char: Start of the line below the block of interest (excluded).
        Defaults to '#', i.e., a comment line.

    :return: List of all the lines that are in between the delimiting ones.
    """
    in_block = False
    data_block = []
    for line in data:
        if in_block:
            if line[: len(bott_char)] != bott_char:
                data_block.append(line)
            else:
                break

        # check for block starter, this is at the end since this line is excluded
        if line[: len(top_char)] == top_char:
            in_block = True

    return data_block



"""Utilities for CRD processors. Mostly methods that can be jitted."""

from typing import List, Tuple, Union

from numba import njit
import numpy as np
from scipy import optimize

from .utilities import fitting, ini, utils


@njit
def create_packages(
    shots: int,
    tofs_mapper: np.ndarray,
    all_tofs: np.ndarray,
) -> Tuple[np.ndarray, np.ndarray]:  # pragma: nocover
    """Create packages from data.

    :param shots: Number of shots per package
    :param tofs_mapper: mapper for ions_per_shot to tofs
    :param all_tofs: all arrival times / bins of ions

    :return: Data array where each row is a full spectrum, each line a package and a
        shot array on how many shots are there per pkg
    """
    bin_start = all_tofs.min()
    bin_end = all_tofs.max()

    nof_pkgs = len(tofs_mapper) // shots
    nof_shots_last_pkg = len(tofs_mapper) % shots
    if nof_shots_last_pkg > 0:
        nof_pkgs += 1

    # number of shots per package
    nof_shots_pkg = np.zeros(nof_pkgs) + shots
    if nof_shots_last_pkg != 0:
        nof_shots_pkg[-1] = nof_shots_last_pkg

    pkg_data = np.zeros((nof_pkgs, bin_end - bin_start + 1))
    for it, tof_map in enumerate(tofs_mapper):
        pkg_it = it // shots
        ions = all_tofs[tof_map[0] : tof_map[1]]
        for ion in ions:
            pkg_data[pkg_it][ion - bin_start] += 1

    return pkg_data, nof_shots_pkg


@njit
def dead_time_correction(
    data: np.ndarray, nof_shots: np.ndarray, dbins: int
) -> np.ndarray:  # pragma: nocover
    """Calculate dead time for a given spectrum.

    :param data: Data array, histogram in bins. 2D array (even for 1D data!)
    :param nof_shots: Number of shots, 1D array of data
    :param dbins: Number of dead bins after original bin (total - 1).

    :return: Dead time corrected data array.
    """
    dbins += 1  # to get total bins

    for lit in range(len(data)):
        ndash = np.zeros(len(data[lit]))  # initialize array to correct with later
        for it in range(len(ndash)):
            # create how far the sum should go
            if it < dbins:
                k = it
            else:
                k = dbins - 1
            # now calculate the sum
            sum_tmp = 0
            for jt in range(k):
                sum_tmp += data[lit][it - (jt + 1)]
            # calculate and add ndash
            ndash[it] = nof_shots[lit] - sum_tmp
        # correct the data
        for it in range(len(data[lit])):
            data[lit][it] = -nof_shots[lit] * np.log(1 - data[lit][it] / ndash[it])

    return data


def delta_calc(names: List[str], integrals: np.ndarray) -> np.ndarray:
    """Calculate delta values for a given set of integrals.

    Use ``iniabu`` to calculate the delta values with respect to normalizing isotope.
    If the name of a peak is not valid or the major isotope not present, return
    ``np.nan`` for that entry. Appropriate error propagation is done as well.

    :param names: Names of the peaks as list.
    :param integrals: Integrals, as defined in ``CRDFileProcessor.integrals``.

    :return: List of delta values, same shape and format as ``integrals``.
    """
    # transform all names to valid ``iniabu`` names or call them ``None``
    names_iniabu = []
    for name in names:
        try:
            names_iniabu.append(ini.iso[name].name)
        except IndexError:
            names_iniabu.append(None)

    # find major isotope names
    norm_iso_name = []
    for name in names_iniabu:
        if name is None:
            norm_iso_name.append(None)
        else:
            ele = name.split("-")[0]
            maj = ini._get_norm_iso(ele)  # can't give index error if above passed
            norm_iso_name.append(maj)

    integrals_dict = dict(zip(names_iniabu, range(len(names_iniabu))))

    integrals_delta = np.zeros_like(integrals, dtype=float)

    for it, iso in enumerate(names_iniabu):
        norm_iso = norm_iso_name[it]

        if iso is None or norm_iso not in names_iniabu:
            integrals_delta[it][0] = np.nan
            integrals_delta[it][1] = np.nan
        else:
            msr_nom = integrals[it][0]
            msr_nom_unc = integrals[it][1]
            msr_denom = integrals[integrals_dict[norm_iso]][0]
            msr_denom_unc = integrals[integrals_dict[norm_iso]][1]

            msr_ratio = msr_nom / msr_denom
            integrals_delta[it][0] = ini.iso_delta(iso, norm_iso, msr_ratio)

            # error calculation
            std_ratio = ini.iso_ratio(iso, norm_iso)
            integrals_delta[it][1] = (
                1000
                / std_ratio
                * np.sqrt(
                    (msr_nom_unc / msr_denom) ** 2
                    + (msr_nom * msr_denom_unc / msr_denom**2) ** 2
                )
            )

    return integrals_delta


def gaussian_fit_get_max(xdata: np.ndarray, ydata: np.ndarray) -> float:
    """Fit a Gaussian to xdata and ydata and return the xvalue of the peak.

    :param xdata: X-axis data
    :param ydata: Y-axis data

    :return: Maximum mof the peak on the x-axis
    """
    mu = xdata[ydata.argmax()]
    sigma = (xdata[-1] - xdata[0]) / 6  # guess
    height = ydata.max()

    coeffs = np.array([mu, sigma, height])

    # need some more error checking here to make sure there really is a peak

    params = optimize.leastsq(fitting.residuals_gaussian, coeffs, args=(ydata, xdata))
    return params[0][0]


def integrals_bg_corr(
    integrals: np.ndarray,
    int_names: np.ndarray,
    int_ch: np.ndarray,
    bgs: np.ndarray,
    bgs_names: np.ndarray,
    bgs_ch: np.ndarray,
    int_pkg: np.ndarray = None,
    bgs_pkg: np.ndarray = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Calculate background correction for integrals with given backgrounds.

    This takes the integrals that already exist and updates them by subtracting the
    backgrounds. Multiple backgrounds per integral can be defined. Important is that the
    names of the backgrounds are equal to the names of the integrals that they
    need to be subtracted from and that the names of the integrals are unique. The
    latter point is tested when defining the integrals.

    .. note:: This routine currently cannot be jitted since we are using an
        ``np.where`` statement. If required for speed, we can go an replace that
        statement. Most likely, this is plenty fast enough though.

    :param integrals: Integrals and uncertianties for all defined peaks.
    :param int_names: Name of the individual peaks. Must be unique values!
    :param int_ch: Number of channels for the whole peak width.
    :param bgs: Backgrounds and their uncertianties for all defined backgrounds.
    :param bgs_names: Peaks each backgrounds go with, can be multiple.
    :param bgs_ch: Number of channels for background width.
    :param int_pkg: Packaged integrals, if exist: otherwise provide ``None``
    :param bgs_pkg: Packaged backgrounds, if exist: otherwise provide ``None``

    :return: Corrected data and data_packages.
    """
    integrals_corr = np.zeros_like(integrals)
    if int_pkg is None:
        integrals_corr_pkg = None
    else:
        integrals_corr_pkg = np.zeros_like(int_pkg)

    def do_correction(
        integrals_in,
        int_names_in,
        int_ch_in,
        bgs_in,
        bgs_names_in,
        bgs_ch_in,
    ):
        """Run the correction, same variable names as outer scope."""
        integrals_corr_in = np.zeros_like(integrals_in)

        bgs_cnt = bgs_in[:, 0]  # get only the counts in the backgrounds, no uncertainty
        bgs_norm = bgs_cnt / bgs_ch_in
        bgs_norm_unc = np.sqrt(bgs_cnt) / bgs_ch_in

        for it in range(len(integrals_in)):
            int_value = integrals_in[it][0]
            bg_indexes = np.where(bgs_names_in == int_names_in[it])[0]
            if len(bg_indexes) > 0:  # background actually exists
                bg_norm = np.sum(bgs_norm[bg_indexes]) / len(bg_indexes)
                bg_norm_unc = np.sum(bgs_norm_unc[bg_indexes]) / len(bg_indexes)

                # write out the corrected values
                integrals_corr_in[it][0] = int_value - int_ch_in[it] * bg_norm
                integrals_corr_in[it][1] = np.sqrt(
                    int_value + bg_norm_unc**2
                )  # sqrt stat, assumes integral uncertainty is sqrt(integral)
            else:
                integrals_corr_in[it][0] = int_value
                integrals_corr_in[it][1] = np.sqrt(int_value)
        return integrals_corr_in

    # for integrals, not packages
    integrals_corr = do_correction(integrals, int_names, int_ch, bgs, bgs_names, bgs_ch)

    if integrals_corr_pkg is not None:
        for it_pkg in range(len(integrals_corr_pkg)):
            integrals_corr_pkg[it_pkg] = do_correction(
                int_pkg[it_pkg], int_names, int_ch, bgs_pkg[it_pkg], bgs_names, bgs_ch
            )

    return integrals_corr, integrals_corr_pkg


@njit
def integrals_summing(
    data: np.ndarray, windows: Tuple[np.ndarray], data_pkg: np.ndarray = None
) -> Tuple[np.ndarray, np.ndarray]:  # pragma: nocover
    """Sum up the integrals within the defined windows and return them.

    :param data: Data to be summed over.
    :param windows: The windows to be investigated (using numpy views)
    :param data_pkg: Package data (optional), if present.

    :return: integrals for data, integrals for data_pkg
    """
    integrals = np.zeros((len(windows), 2))

    # packages
    integrals_pkg = None
    if data_pkg is not None:
        integrals_pkg = np.zeros((data_pkg.shape[0], len(windows), 2))
        for ht in range(len(data_pkg)):
            for it, window in enumerate(windows):
                integrals_pkg[ht][it][0] = data_pkg[ht][window].sum()
                integrals_pkg[ht][it][1] = np.sqrt(integrals_pkg[ht][it][0])
        # define all integrals as the sum of the packages -> allow for filtering
        integrals[:, 0] = integrals_pkg.sum(axis=0)[:, 0]
        integrals[:, 1] = np.sqrt(np.sum(integrals_pkg[:, :, 1] ** 2, axis=0))
    else:
        for it, window in enumerate(windows):
            integrals[it][0] = data[window].sum()
            integrals[it][1] = np.sqrt(integrals[it][0])

    return integrals, integrals_pkg


@njit
def mask_filter_max_ions_per_time(
    ions_per_shot: np.array,
    tofs: np.array,
    max_ions: int,
    time_chan: int,
) -> np.array:  # pragma: nocover
    """Return indices where more than wanted shots are in a time window.

    :param ions_per_shot: How many ions are there per shot? Also defines the shape of
        the return array.
    :param tofs: All ToFs. Must be of length ions_per_shot.sum().
    :param max_ions: Maximum number of ions that are allowed in channel window.
    :param time_chan: Width of the window in channels (bins).

    :return: Boolean array of shape like ions_per_shot if more are in or not.
    """
    return_mask = np.zeros_like(ions_per_shot)  # initialize return mask

    start_ind = 0

    for it, ips in enumerate(ions_per_shot):
        end_ind = start_ind + ips
        tofs_shot = tofs[start_ind:end_ind]

        # run the filter
        for tof in tofs_shot:
            tofs_diff = np.abs(tofs_shot - tof)  # differences
            ions_in_window = len(
                np.where(tofs_diff <= time_chan)[0]
            )  # where diff small
            if ions_in_window > max_ions:  # comparison with max allowed
                return_mask[it] = 1
                break  # break this for loop: one true is enough to kick the shot

        start_ind = end_ind

    return np.where(return_mask == 1)[0]


@njit
def mask_filter_max_ions_per_tof_window(
    ions_per_shot: np.array,
    tofs: np.array,
    max_ions: int,
    tof_window: np.array,
) -> np.array:  # pragma: nocover
    """Return indices where more than wanted shots are in a given ToF window.

    :param ions_per_shot: How many ions are there per shot? Also defines the shape of
        the return array.
    :param tofs: All ToFs. Must be of length ions_per_shot.sum().
    :param max_ions: Maximum number of ions that are allowed in channel window.
    :param tof_window: Start and stop time of the ToF window in channel numbers.

    :return: Boolean array of shape like ions_per_shot if more are in or not.
    """
    return_mask = np.zeros_like(ions_per_shot)  # initialize return mask

    start_ind = 0

    for it, ips in enumerate(ions_per_shot):
        end_ind = start_ind + ips
        tofs_shot = tofs[start_ind:end_ind]

        if tofs_shot.shape[0] == 0:
            continue

        filtered_tofs = np.where(
            np.logical_and(tofs_shot >= tof_window[0], tofs_shot <= tof_window[1])
        )[0]
        nof_tofs_win = len(filtered_tofs)

        if nof_tofs_win > max_ions:
            return_mask[it] = 1

        start_ind = end_ind

    return np.where(return_mask == 1)[0]


def mass_calibration(
    params: np.array, tof: np.array, return_params: bool = False
) -> Union[np.array, Tuple[np.array]]:
    """Perform the mass calibration.

    :param params: Parameters for mass calibration.
    :param tof: Array with all the ToFs that need a mass equivalent.
    :param return_params: Return parameters as well? Defaults to False

    :return: Mass for given ToF.
    """
    # function to return mass with a given functional form
    calc_mass = tof_to_mass

    # calculate the initial guess for scipy fitting routine
    ch1 = params[0][0]
    m1 = params[0][1]
    ch2 = params[1][0]
    m2 = params[1][1]
    t0 = (ch1 * np.sqrt(m2) - ch2 * np.sqrt(m1)) / (np.sqrt(m2) - np.sqrt(m1))
    b = np.sqrt((ch1 - t0) ** 2.0 / m1)

    # fit the curve and store the parameters
    params_fit = optimize.curve_fit(calc_mass, params[:, 0], params[:, 1], p0=(t0, b))

    mass = calc_mass(tof, params_fit[0][0], params_fit[0][1])

    if return_params:
        return mass, params_fit[0]
    else:
        return mass


def mass_to_tof(
    m: Union[np.ndarray, float], tm0: float, const: float
) -> Union[np.ndarray, float]:
    r"""Functional prescription to turn mass into ToF.

    Returns the ToF with the defined functional description for a mass calibration.
    Two parameters are required. The equation, with parameters defined as below,
    is as following:

    .. math:: t = \sqrt{m} \cdot \mathrm{const} + t_{0}

    :param m: mass
    :param tm0: parameter 1
    :param const: parameter 2

    :return: time
    """
    return np.sqrt(m) * const + tm0


def multi_range_indexes(rng: np.array) -> np.array:
    """Create multi range indexes.

    If a range is given as (from, to), the from will be included, while the to will
    be excluded.

    :param rng: Range, given as a numpy array of two entries each.

    :return: A 1D array with all the indexes spelled out. This allows for viewing
        numpy arrays for multiple windows.
    """
    num_shots = 0
    ind_tmp = []
    for rit in rng:
        if rit[0] != rit[1]:
            arranged_tmp = np.arange(rit[0], rit[1])
            ind_tmp.append(arranged_tmp)
            num_shots += len(arranged_tmp)

    indexes = np.zeros(num_shots, dtype=int)
    ind_b = 0
    for rit in ind_tmp:
        ind_e = ind_b + len(rit)
        indexes[ind_b:ind_e] = rit
        ind_b = ind_e
    return indexes


@njit
def remove_shots_from_filtered_packages_ind(
    shots_rejected: np.array,
    len_indexes: int,
    filtered_pkg_ind: np.array,
    pkg_size: int,
) -> Tuple[np.array, np.array]:  # pragma: nocover
    """Remove packages that were already filtered pkg from ion filter indexes.

    This routine is used to filter indexes in case a package filter has been applied,
    and now an ion / shot based filter needs to be applied.

    :param shots_rejected: Array of indexes with rejected shots.
    :param len_indexes: length of the indexes that the rejected shots are from.
    :param pkg_size: Size of the packages that were created.
    :param filtered_pkg_ind: Array with indexes of packages that have been filtered.

    :return: List of two Arrays with shots_indexes and shots_rejected, but filtered.
    """
    shots_indexes = utils.not_index(shots_rejected, len_indexes)
    for pkg_it in filtered_pkg_ind:
        lower_lim = pkg_it * pkg_size
        upper_lim = lower_lim + pkg_size
        shots_indexes = shots_indexes[
            np.where(
                np.logical_or(shots_indexes < lower_lim, shots_indexes >= upper_lim)
            )
        ]
        shots_rejected = shots_rejected[
            np.where(
                np.logical_or(shots_rejected < lower_lim, shots_rejected >= upper_lim)
            )
        ]
    return shots_indexes, shots_rejected


@njit
def remove_shots_from_packages(
    pkg_size: int,
    shots_rejected: np.array,
    ions_to_tof_map: np.array,
    all_tofs: np.array,
    data_pkg: np.array,
    nof_shots_pkg: np.array,
    pkg_filtered_ind: np.array = None,
) -> Tuple[np.array, np.array]:  # pragma: nocover
    """Remove shots from packages.

    This routine can take a list of individual ions and remove them from fully
    packaged data. In addition, it can also take a list of packages that, with respect
    to the raw data, have previously been removed. This is useful in order to filter
    individual shots from packages after packages themselves have been filtered.

    :param pkg_size: How many shots were grouped into a package originally?
    :param shots_rejected: Index array of the rejected shots.
    :param ions_to_tof_map: Mapping array where ions are in all_tof array.
    :param all_tofs: Array containing all the ToFs.
    :param data_pkg: Original data_pkg before filtering.
    :param nof_shots_pkg: Original nof_shots_pkg before filtering.
    :param pkg_filtered_ind: Indexes where the filtered packages are.

    :return: Filtered data_pkg and nof_shots_pkg arrays.
    """
    for shot_rej in shots_rejected:
        # calculate index of package
        pkg_ind = shot_rej // pkg_size

        if pkg_filtered_ind is not None:
            # need to subtract number of filtered packages up to here!
            pkg_rej_until = len(np.where(pkg_filtered_ind < pkg_ind))
            pkg_ind -= pkg_rej_until

        # get tofs to subtract from package and set up array with proper sizes
        rng_tofs = ions_to_tof_map[shot_rej]
        ions_to_sub = all_tofs[rng_tofs[0] : rng_tofs[1]]
        array_to_sub = np.zeros_like(data_pkg[pkg_ind])
        array_to_sub[ions_to_sub - all_tofs.min()] += 1

        data_pkg[pkg_ind] -= array_to_sub
        nof_shots_pkg[pkg_ind] -= 1

        return data_pkg, nof_shots_pkg


@njit
def sort_data_into_spectrum(
    ions: np.ndarray, bin_start: int, bin_end: int
) -> np.ndarray:  # pragma: nocover
    """Sort ion data in 1D array into an overall array and sum them up.

    :param ions: Arrival time of the ions - number of time bin
    :param bin_start: First bin of spectrum
    :param bin_end: Last bin of spectrum

    :return: arrival bins summed up
    """
    data = np.zeros(bin_end - bin_start + 1)
    for ion in ions:
        data[ion - bin_start] += 1
    return data


def tof_to_mass(
    tm: Union[np.ndarray, float], tm0: float, const: float
) -> Union[np.ndarray, float]:
    r"""Functional prescription to turn ToF into mass.

    Returns the mass with the defined functional description for a mass calibration.
    Two parameters are required. The equation, with parameters defined as below,
    is as following:

    .. math:: m = \left(\frac{tm - tm_{0}}{\mathrm{const}}\right)^{2}

    :param tm: time or channel
    :param tm0: parameter 1
    :param const: parameter 2

    :return: mass m
    """
    return ((tm - tm0) / const) ** 2



"""Unit tests for legacy file reader."""

from pathlib import Path

import pytest
import numpy as np

from rimseval.compatibility.lion_eval import LIONEvalCal


def test_lion_eval_cal_file_type_error():
    """Raise TypeError if not initialized with a path."""
    with pytest.raises(TypeError):
        _ = LIONEvalCal(13)


def test_lion_eval_cal_file_all(legacy_files_path):
    """Read an existing calibration file with all information."""
    fname = Path(legacy_files_path).joinpath("all_cals.cal")
    cal = LIONEvalCal(fname)
    assert isinstance(cal.mass_cal, np.ndarray)
    assert isinstance(cal.integrals, list)
    assert isinstance(cal.bg_corr, list)


def test_lion_eval_cal_file_mcal_only(legacy_files_path):
    """Read an existing calibration file with only mass calibration."""
    fname = Path(legacy_files_path).joinpath("mcal_only.cal")
    cal = LIONEvalCal(fname)
    assert isinstance(cal.mass_cal, np.ndarray)
    assert cal.integrals is None
    assert cal.bg_corr is None


def test_lion_eval_cal_file_no_cal(legacy_files_path):
    """Read an existing calibration file with no calibration information."""
    fname = Path(legacy_files_path).joinpath("no_cal.cal")
    cal = LIONEvalCal(fname)
    assert cal.mass_cal is None
    assert cal.integrals is None
    assert cal.bg_corr is None


def test_lion_eval_applied_filters(legacy_files_path):
    """Set applied filters data from `lioneval_full_calfile`."""
    fname = Path(legacy_files_path).joinpath("lioneval_full_calfile.cal")
    cal = LIONEvalCal(fname)

    dict_exp = {
        "dead_time_corr": [True, 7],
        "packages": [True, 1000],
        "max_ions_per_shot": [True, 10],
        "max_ions_per_pkg": [True, 1000],
        "max_ions_per_time": [True, 50, float(200)],
        "max_ions_per_tof_window": [True, 100, [float(5), float(8)]],
        "spectrum_part": [True, [1, 70000]],
    }

    assert cal.applied_filters == dict_exp


def test_lion_eval_applied_filters_no_block(legacy_files_path):
    """No applied filters if no block given."""
    fname = Path(legacy_files_path).joinpath("lioneval_full_calfile.cal")
    cal = LIONEvalCal(fname)
    assert cal._read_and_set_settings_calculation(None) is None


def test_lion_eval_applied_filters_bad_settings(legacy_files_path):
    """Applied filters stays ``None`` if bad calibration file."""
    fname = Path(legacy_files_path).joinpath("bad_settings.cal")
    cal = LIONEvalCal(fname)
    assert cal.applied_filters is None



"""GUIs to interactively set various variables in package."""

from .integrals import define_backgrounds_app, define_integrals_app
from .mcal import create_mass_cal_app
from .plots import dt_ions, integrals_packages, nof_ions_per_shot

__all__ = [
    "create_mass_cal_app",
    "define_backgrounds_app",
    "define_integrals_app",
    "dt_ions",
    "integrals_packages",
    "nof_ions_per_shot",
]



"""Example: Re-implementation of maximum ions per package filter."""

import numpy as np

from rimseval.processor import CRDFileProcessor


def calc(crd: CRDFileProcessor) -> None:
    """Macro to filter out all packages with more than 4 ions.

    Here we assume, that you have already created packages using the built-in function.
    Package data is therefore available.

    :param crd: CRD file processor that will be passed to the macro.
    """
    # maximum ions per pkg to be filtered out
    max_ions = 4

    # calculate total ions in all the packages
    total_ions_per_pkg = np.sum(crd.data_pkg, axis=1)

    # find packages that need deletion
    pkg_to_delete = np.where(total_ions_per_pkg > max_ions)[0]

    # now delete the packages that are not needed
    crd.data_pkg = np.delete(crd.data_pkg, pkg_to_delete, axis=0)

    # update the number of shots per package array: delete entries of deleted packages
    crd.nof_shots_pkg = np.delete(crd.nof_shots_pkg, pkg_to_delete, axis=0)

    # finally, we need to update the ``crd.data`` and ``crd.nof_shots`` entries
    crd.data = np.sum(crd.data_pkg, axis=0)
    crd.nof_shots = np.sum(crd.nof_shots_pkg)



"""Unit tests for lst to crd converter."""

from pathlib import Path

import pytest

from rimseval.data_io import LST2CRD


def test_binary_data_not_implemented(lst_crd_path):
    """Raise NotImplementedData when a binary file is loaded."""
    lst_fname = "mcs8a_binary_short_100k_signal_dg.lst"

    err_msg_exp = "Binary data is currently not supported."

    with pytest.raises(NotImplementedError) as err:
        conv = LST2CRD(lst_crd_path.joinpath(lst_fname), channel_data=7)
        conv.read_list_file()

    msg = err.value.args[0]
    assert msg == err_msg_exp


def test_instrument_not_implemented_error(lst_crd_path):
    """Raise NotImplementedError if instrument in lst file is invalid."""
    lst_fname = "err_invalid_instrument.lst"

    err_msg_incl = "[INVALID_INSTRUMENT A]"

    conv = LST2CRD(lst_crd_path.joinpath(lst_fname), channel_data=7)
    with pytest.raises(NotImplementedError) as err:
        conv.read_list_file()

    msg = err.value.args[-1]
    assert err_msg_incl in msg


@pytest.mark.parametrize("tag_ch", [None, 3])
def test_ions_out_of_range_warning(tag_ch, tmpdir, lst_crd_path):
    """Raise a warning if ions are out of range."""
    lst_fname = "err_ions_out_of_range.lst"
    # copy file to temporary file
    tmpdir.join(lst_fname).write_binary(lst_crd_path.joinpath(lst_fname).read_bytes())
    lst_fpath = Path(tmpdir.strpath).joinpath(lst_fname)
    # create crd
    conv = LST2CRD(file_name=lst_fpath, channel_data=4, channel_tag=tag_ch)
    conv.read_list_file()

    with pytest.warns(
        UserWarning,
        match="lst file contained ions that were outside the allowed range.",
    ):
        conv.write_crd()


def test_mcs6a_short_10k(tmpdir, lst_crd_path):
    """Convert MCS6A file to crd and compare outputs."""
    lst_fname = "MCS6A_short_10k_signal.lst"
    crd_fname = "MCS6A_short_10k_signal.crd"
    # copy file to temporary file
    tmpdir.join(lst_fname).write_binary(lst_crd_path.joinpath(lst_fname).read_bytes())
    lst_fpath = Path(tmpdir.strpath).joinpath(lst_fname)
    # create crd
    conv = LST2CRD(file_name=lst_fpath, channel_data=4)
    conv.read_list_file()
    conv.write_crd()
    # compare crd
    assert (
        Path(tmpdir.strpath).joinpath(crd_fname).read_bytes()
        == lst_crd_path.joinpath(crd_fname).read_bytes()
    )


def test_mcs6a_short_10k_tagged(tmpdir, lst_crd_path):
    """Convert tagged MCS6A file to two crd files and compare outputs."""
    lst_fname = "MCS6A_short_10k_signal_tag.lst"
    crd_fname_tagged = "MCS6A_short_10k_signal_tag.tagged.crd"
    crd_fname_untagged = "MCS6A_short_10k_signal_tag.untagged.crd"
    # copy file to temporary file
    tmpdir.join(lst_fname).write_binary(lst_crd_path.joinpath(lst_fname).read_bytes())
    lst_fpath = Path(tmpdir.strpath).joinpath(lst_fname)
    # create crd
    conv = LST2CRD(file_name=lst_fpath, channel_data=4, channel_tag=3)
    conv.read_list_file()
    conv.write_crd()
    # compare crds
    assert (
        Path(tmpdir.strpath).joinpath(crd_fname_tagged).read_bytes()
        == lst_crd_path.joinpath(crd_fname_tagged).read_bytes()
    )
    assert (
        Path(tmpdir.strpath).joinpath(crd_fname_untagged).read_bytes()
        == lst_crd_path.joinpath(crd_fname_untagged).read_bytes()
    )


def test_mcs8a_short_10k(tmpdir, lst_crd_path):
    """Convert MCS8a file to crd and compare outputs."""
    lst_fname = "MCS8a_short_10k_signal.lst"
    crd_fname = "MCS8a_short_10k_signal.crd"
    # copy file to temporary file
    tmpdir.join(lst_fname).write_binary(lst_crd_path.joinpath(lst_fname).read_bytes())
    lst_fpath = Path(tmpdir.strpath).joinpath(lst_fname)
    # create crd
    conv = LST2CRD(file_name=lst_fpath, channel_data=9)
    conv.read_list_file()
    conv.write_crd()
    # compare crd
    assert (
        Path(tmpdir.strpath).joinpath(crd_fname).read_bytes()
        == lst_crd_path.joinpath(crd_fname).read_bytes()
    )


def test_mcs8a_short_10k_wrong_channel_error_message(tmpdir, lst_crd_path):
    """Raise OSError and propose the correct channel to user"""
    lst_fname = "MCS8a_short_10k_signal.lst"
    other_channels = [9]

    # copy file to temporary file
    tmpdir.join(lst_fname).write_binary(lst_crd_path.joinpath(lst_fname).read_bytes())
    lst_fpath = Path(tmpdir.strpath).joinpath(lst_fname)

    # create crd
    conv = LST2CRD(file_name=lst_fpath, channel_data=7)
    conv.read_list_file()
    with pytest.raises(OSError) as err:
        conv.write_crd()
    msg = err.value.args[0]
    assert msg == (
        f"There are no counts present in this file. Please double "
        f"check that you are using the correct channel for the signal. "
        f"The file seems to have counts in channels {other_channels}."
    )


def test_no_data_type_error(lst_crd_path):
    """Raise OSError if file contains no data type."""
    lst_fname = "err_no_datatype.lst"

    err_msg_exp = "Could not find a data type in the list file!"

    conv = LST2CRD(lst_crd_path.joinpath(lst_fname), channel_data=7)
    with pytest.raises(OSError) as err:
        conv.read_list_file()

    msg = err.value.args[-1]
    assert err_msg_exp in msg



"""This file contains utilities for processing list files."""

from typing import List, Tuple

from numba import njit
import numpy as np

from .lst_to_crd import LST2CRD


def ascii_to_ndarray(
    data_list: List[str], fmt: LST2CRD.ASCIIFormat, channel: int, tag: int = None
) -> Tuple[np.ndarray, np.ndarray, List]:
    """Turn ASCII LST data to a numpy array.

    Takes the whole data block and returns the data in a properly formatted numpy array.
    If channels other than the selected ones are available, these are written to a
    List and also returned as ``other_channels``.

    :param data_list: Data, directly supplied from the TDC block.
    :param fmt: Format of the data
    :param channel: Channel the data is in
    :param tag: Channel the tag is in, or None if no tag

    :return: Data, Tag Data, Other Channels available
    """
    # prepare the data and list
    data_arr = np.empty((len(data_list), 2), dtype=np.uint32)
    data_arr_tag = None
    # initalize stuff for tags
    if tag is not None:
        data_arr_tag = np.empty(
            len(data_list), dtype=np.uint32
        )  # only sweep, not the channel

    # some helper variables for easy conversion
    binary_width = fmt.value[0]
    boundaries = fmt.value[1]

    # counter for ions in the right channel
    ion_counter = 0

    tag_counter = 0

    other_channels = []

    # transform to bin number with correct length
    for data in data_list:
        if data != "":
            bin_tmp = f"{int(data, 16):{binary_width}b}".replace(" ", "0")
            # parse data
            tmp_channel = int(bin_tmp[boundaries[2][0] : boundaries[2][1]], 2)
            if tmp_channel == channel:
                swp_val, time_val = get_sweep_time_ascii(
                    bin_tmp, boundaries[0], boundaries[1]
                )
                data_arr[ion_counter][0] = swp_val
                data_arr[ion_counter][1] = time_val
                ion_counter += 1
            elif tmp_channel == tag:
                swp_val, _ = get_sweep_time_ascii(bin_tmp, boundaries[0], boundaries[1])
                data_arr_tag[tag_counter] = swp_val
                tag_counter += 1
            elif tmp_channel != 0:
                if tmp_channel not in other_channels:
                    other_channels.append(tmp_channel)

    data_arr = data_arr[:ion_counter]
    if tag is not None:
        data_arr_tag = data_arr_tag[:tag_counter]

    return data_arr, data_arr_tag, other_channels


def get_sweep_time_ascii(
    data: str, sweep_b: Tuple[int, int], time_b: Tuple[int, int]
) -> Tuple[int, int]:
    """Get sweep and time from a given ASCII string.

    :param data: ASCII string
    :param sweep_b: Boundaries of sweep
    :param time_b: Boundaries of time

    :return: sweep, time
    """
    sweep_val = int(data[sweep_b[0] : sweep_b[1]], 2)
    time_val = int(data[time_b[0] : time_b[1]], 2)
    return sweep_val, time_val


@njit
def transfer_lst_to_crd_data(
    data_in: np.ndarray, max_sweep: int, ion_range: int
) -> Tuple[np.ndarray, np.ndarray, bool]:  # pragma: nocover
    """Transfer lst file specific data to the crd format.

    :param data_in: Array: One ion per line, two entries: sweep first (shot), then time
    :param max_sweep: the maximum sweep that can be represented by data resolution
    :param ion_range: Valid range of the data in multiples of 100ps bins

    :return: Array of how many ions are in each shot, Array of all arrival times of
        these ions, and a bool if there are any ions out of range
    """
    data = data_in.copy()

    # go through and sort out max range issues
    threshold = max_sweep // 2
    multiplier = 0
    last_shot = data[0][0]
    for it in range(1, data.shape[0]):
        curr_shot = data[it][0]
        if (
            curr_shot < threshold < last_shot and last_shot - curr_shot > threshold
        ):  # need to flip forward
            multiplier += 1
        elif (
            last_shot < threshold < curr_shot and curr_shot - last_shot > threshold
        ):  # flip back
            multiplier -= 1
        # modify data
        adder = multiplier * max_sweep
        data[it][0] += adder
        last_shot = curr_shot

    # now sort the np array
    data_sort = data[data[:, 0].argsort()]

    # now create the shots and ions arrays and fill them
    shots = np.zeros(data_sort[:, 0].max(), dtype=np.uint32)
    ions = np.empty(
        len(data_sort[:, 1][np.where(data_sort[:, 1] <= ion_range)]), dtype=np.uint32
    )

    it = 0
    ions_out_of_range = False
    for shot, ion in data_sort:
        if ion <= ion_range:
            shots[shot - 1] += 1  # zero versus one based
            ions[it] = ion
            it += 1
        else:
            ions_out_of_range = True

    return shots, ions, ions_out_of_range


@njit
def separate_signal_with_tag(
    data_arr: np.ndarray, tag_arr: np.ndarray
) -> Tuple[np.ndarray, np.ndarray]:  # pragma: nocover
    """Separate a dataset into a tagged and an untagged data set.

    :param data_arr: Array with all the data, shot, bins.
    :param tag_arr: 1d array with shots that are tagged.

    :return: Tagged and untagged data (split) in same format as ``data_arr``
    """
    tmp_untagged = np.empty_like(data_arr)
    tmp_tagged = np.empty_like(data_arr)

    cnt_untagged = 0
    cnt_tagged = 0

    for dat in data_arr:
        if dat[0] in tag_arr:  # tagged data
            tmp_tagged[cnt_tagged] = dat
            cnt_tagged += 1
        else:  # untagged data
            tmp_untagged[cnt_untagged] = dat
            cnt_untagged += 1

    data_untagged = tmp_untagged[:cnt_untagged]
    data_tagged = tmp_tagged[:cnt_tagged]

    return data_untagged, data_tagged
